#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Volume v3 snapshot action implementations"""

import argparse
from collections.abc import Iterable, Sequence
import functools
import logging
from typing import Any

from cliff import columns as cliff_columns
from openstack.block_storage.v3 import snapshot as _snapshot
from openstack import utils as sdk_utils
from osc_lib.cli import format_columns
from osc_lib.cli import parseractions
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.common import pagination
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common

LOG = logging.getLogger(__name__)


class VolumeIdColumn(cliff_columns.FormattableColumn[str]):
    """Formattable column for volume ID column.

    Unlike the parent FormattableColumn class, the initializer of the
    class takes volume_cache as the second argument.
    osc_lib.utils.get_item_properties instantiate cliff FormattableColumn
    object with a single parameter "column value", so you need to pass
    a partially initialized class like
    ``functools.partial(VolumeIdColumn, volume_cache)``.
    """

    def __init__(
        self, value: str, volume_cache: dict[str, Any] | None = None
    ) -> None:
        super().__init__(value)
        self._volume_cache = volume_cache or {}

    def human_readable(self) -> str:
        """Return a volume name if available

        :rtype: either the volume ID or name
        """
        volume_id = self._value
        volume = volume_id
        if volume_id in self._volume_cache.keys():
            volume = self._volume_cache[volume_id].name
        return volume


def _format_snapshot(snapshot: _snapshot.Snapshot) -> dict[str, object]:
    # Some columns returned by openstacksdk should not be shown because they're
    # either irrelevant or duplicates
    ignored_columns = {
        # computed columns
        'location',
        # create-only columns
        'consumes_quota',
        'force',
        'group_snapshot_id',
        # ignored columns
        'os-extended-snapshot-attributes:progress',
        'os-extended-snapshot-attributes:project_id',
        'updated_at',
        'user_id',
        # unnecessary columns
        'links',
    }

    info = snapshot.to_dict(original_names=True)
    data = {}
    for key, value in info.items():
        if key in ignored_columns:
            continue

        data[key] = value

    data.update(
        {
            'properties': format_columns.DictColumn(data.pop('metadata')),
        }
    )

    return data


class CreateVolumeSnapshot(command.ShowOne):
    _description = _("Create new volume snapshot")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "snapshot_name",
            metavar="<snapshot-name>",
            help=_("Name of the new snapshot"),
        )
        parser.add_argument(
            "--volume",
            metavar="<volume>",
            help=_(
                "Volume to snapshot (name or ID) (default is <snapshot-name>)"
            ),
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            help=_("Description of the snapshot"),
        )
        parser.add_argument(
            "--force",
            action="store_true",
            default=False,
            help=_(
                "Allow snapshot of in-use (attached) volume. "
                "Only needed for microversions prior to 3.66; "
                "ignored for 3.66+"
            ),
        )
        parser.add_argument(
            "--property",
            metavar="<key=value>",
            dest='properties',
            action=parseractions.KeyValueAction,
            help=_(
                "Set a property to this snapshot "
                "(repeat option to set multiple properties)"
            ),
        )
        parser.add_argument(
            "--remote-source",
            metavar="<key=value>",
            action=parseractions.KeyValueAction,
            help=_(
                "The attribute(s) of the existing remote volume snapshot "
                "(admin required) (repeat option to specify multiple "
                "attributes) e.g.: '--remote-source source-name=test_name "
                "--remote-source source-id=test_id'"
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        volume = parsed_args.volume
        if not parsed_args.volume:
            volume = parsed_args.snapshot_name
        volume_id = volume_client.find_volume(volume, ignore_missing=False).id

        if parsed_args.remote_source:
            # Create a new snapshot from an existing remote snapshot source
            if parsed_args.force:
                msg = _(
                    "'--force' option will not work when you create "
                    "new volume snapshot from an existing remote "
                    "volume snapshot"
                )
                LOG.warning(msg)

            snapshot = volume_client.manage_snapshot(
                volume_id=volume_id,
                ref=parsed_args.remote_source,
                name=parsed_args.snapshot_name,
                description=parsed_args.description,
                metadata=parsed_args.properties,
            )
        else:
            # Create a new snapshot from scratch
            # only for microversion < 3.66, pass force parameter
            # for backward compatibility
            if not sdk_utils.supports_microversion(volume_client, '3.66'):
                snapshot = volume_client.create_snapshot(
                    volume_id=volume_id,
                    force=parsed_args.force,
                    name=parsed_args.snapshot_name,
                    description=parsed_args.description,
                    metadata=parsed_args.properties,
                )
            else:
                snapshot = volume_client.create_snapshot(
                    volume_id=volume_id,
                    name=parsed_args.snapshot_name,
                    description=parsed_args.description,
                    metadata=parsed_args.properties,
                )

        data = _format_snapshot(snapshot)
        col_headers, col_data = zip(*sorted(data.items()))
        return col_headers, col_data


class DeleteVolumeSnapshot(command.Command):
    _description = _("Delete volume snapshot(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "snapshots",
            metavar="<snapshot>",
            nargs="+",
            help=_("Snapshot(s) to delete (name or ID)"),
        )
        parser.add_argument(
            '--force',
            action='store_true',
            help=_(
                "Attempt forced removal of snapshot(s), "
                "regardless of state (defaults to False)"
            ),
        )
        parser.add_argument(
            '--remote',
            action='store_true',
            help=_(
                'Unmanage the snapshot, removing it from the Block Storage '
                'service management but not from the backend.'
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        result = 0

        if parsed_args.remote:
            if parsed_args.force:
                msg = _(
                    "The --force option is not supported with the "
                    "--remote parameter."
                )
                raise exceptions.CommandError(msg)

        for snapshot in parsed_args.snapshots:
            try:
                snapshot_id = volume_client.find_snapshot(
                    snapshot, ignore_missing=False
                ).id
                if parsed_args.remote:
                    volume_client.unmanage_snapshot(snapshot_id)
                else:
                    volume_client.delete_snapshot(
                        snapshot_id, force=parsed_args.force
                    )
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete snapshot with "
                        "name or ID '%(snapshot)s': %(e)s"
                    ),
                    {'snapshot': snapshot, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.snapshots)
            msg = _("%(result)s of %(total)s snapshots failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListVolumeSnapshot(command.Lister):
    _description = _("List volume snapshots")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--all-projects',
            action='store_true',
            default=False,
            help=_('Include all projects (admin only)'),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_('Filter results by project (name or ID) (admin only)'),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_('List additional fields in output'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            default=None,
            help=_('Filters results by a name.'),
        )
        parser.add_argument(
            '--status',
            metavar='<status>',
            choices=[
                'available',
                'error',
                'creating',
                'deleting',
                'error_deleting',
            ],
            help=_(
                "Filters results by a status. "
                "('available', 'error', 'creating', 'deleting'"
                " or 'error_deleting')"
            ),
        )
        parser.add_argument(
            '--volume',
            metavar='<volume>',
            default=None,
            help=_('Filters results by a volume (name or ID).'),
        )
        pagination.add_marker_pagination_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        columns: tuple[str, ...] = (
            'id',
            'name',
            'description',
            'status',
            'size',
        )
        column_headers: tuple[str, ...] = (
            'ID',
            'Name',
            'Description',
            'Status',
            'Size',
        )
        if parsed_args.long:
            columns += (
                'created_at',
                'volume_id',
                'metadata',
            )
            column_headers += (
                'Created At',
                'Volume',
                'Properties',
            )

        # Cache the volume list
        volume_cache = {}
        try:
            for s in volume_client.volumes():
                volume_cache[s.id] = s
        except Exception:  # noqa: S110
            # Just forget it if there's any trouble
            pass
        _VolumeIdColumn = functools.partial(
            VolumeIdColumn, volume_cache=volume_cache
        )

        volume_id = None
        if parsed_args.volume:
            volume_id = volume_client.find_volume(
                parsed_args.volume, ignore_missing=False
            ).id

        project_id = None
        if parsed_args.project:
            project_id = identity_common.find_project_id_sdk(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            )

        # set value of 'all_tenants' when using project option
        all_projects = (
            True if parsed_args.project else parsed_args.all_projects
        )

        data = volume_client.snapshots(
            marker=parsed_args.marker,
            limit=parsed_args.limit,
            max_items=parsed_args.max_items,
            all_projects=all_projects,
            project_id=project_id,
            name=parsed_args.name,
            status=parsed_args.status,
            volume_id=volume_id,
        )
        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={
                        'metadata': format_columns.DictColumn,
                        'volume_id': _VolumeIdColumn,
                    },
                )
                for s in data
            ),
        )


class SetVolumeSnapshot(command.Command):
    _description = _("Set volume snapshot properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'snapshot',
            metavar='<snapshot>',
            help=_('Snapshot to modify (name or ID)'),
        )
        parser.add_argument(
            '--name', metavar='<name>', help=_('New snapshot name')
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('New snapshot description'),
        )
        parser.add_argument(
            "--no-property",
            dest="no_property",
            action="store_true",
            help=_(
                "Remove all properties from <snapshot> "
                "(specify both --no-property and --property to "
                "remove the current properties before setting "
                "new properties.)"
            ),
        )
        parser.add_argument(
            '--property',
            metavar='<key=value>',
            action=parseractions.KeyValueAction,
            dest='properties',
            help=_(
                'Property to add/change for this snapshot '
                '(repeat option to set multiple properties)'
            ),
        )
        parser.add_argument(
            '--state',
            metavar='<state>',
            choices=[
                'available',
                'error',
                'creating',
                'deleting',
                'error_deleting',
            ],
            help=_(
                'New snapshot state. ("available", "error", "creating", '
                '"deleting", or "error_deleting") (admin only) '
                '(This option simply changes the state of the snapshot '
                'in the database with no regard to actual status, '
                'exercise caution when using)'
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        snapshot = volume_client.find_snapshot(
            parsed_args.snapshot, ignore_missing=False
        )

        result = 0
        if parsed_args.no_property:
            try:
                volume_client.delete_snapshot_metadata(
                    snapshot.id, keys=list(snapshot.metadata)
                )
            except Exception as e:
                LOG.error(_("Failed to clean snapshot properties: %s"), e)
                result += 1

        if parsed_args.properties:
            try:
                volume_client.set_snapshot_metadata(
                    snapshot.id, **parsed_args.properties
                )
            except Exception as e:
                LOG.error(_("Failed to set snapshot property: %s"), e)
                result += 1

        if parsed_args.state:
            try:
                volume_client.reset_snapshot_status(
                    snapshot.id, parsed_args.state
                )
            except Exception as e:
                LOG.error(_("Failed to set snapshot state: %s"), e)
                result += 1

        kwargs = {}
        if parsed_args.name:
            kwargs['name'] = parsed_args.name
        if parsed_args.description:
            kwargs['description'] = parsed_args.description
        if kwargs:
            try:
                volume_client.update_snapshot(snapshot.id, **kwargs)
            except Exception as e:
                LOG.error(
                    _("Failed to update snapshot name or description: %s"),
                    e,
                )
                result += 1

        if result > 0:
            raise exceptions.CommandError(
                _("One or more of the set operations failed")
            )


class ShowVolumeSnapshot(command.ShowOne):
    _description = _("Display volume snapshot details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "snapshot",
            metavar="<snapshot>",
            help=_("Snapshot to display (name or ID)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        snapshot = volume_client.find_snapshot(
            parsed_args.snapshot, ignore_missing=False
        )

        data = _format_snapshot(snapshot)
        col_headers, col_data = zip(*sorted(data.items()))
        return col_headers, col_data


class UnsetVolumeSnapshot(command.Command):
    _description = _("Unset volume snapshot properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'snapshot',
            metavar='<snapshot>',
            help=_('Snapshot to modify (name or ID)'),
        )
        parser.add_argument(
            '--property',
            metavar='<key>',
            dest='properties',
            action='append',
            default=[],
            help=_(
                'Property to remove from snapshot '
                '(repeat option to remove multiple properties)'
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        snapshot = volume_client.find_snapshot(
            parsed_args.snapshot, ignore_missing=False
        )

        if parsed_args.properties:
            volume_client.delete_snapshot_metadata(
                snapshot.id, keys=parsed_args.properties
            )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest.mock import call

from openstack.identity.v3 import project as _project
from openstack.network.v2 import (
    security_groups_default_statefulness as _sg_default_statefulness,
)
from openstack.test import fakes as sdk_fakes
from osc_lib import exceptions

from openstackclient.network.v2 import security_groups_default_statefulness
from openstackclient.tests.unit.network.v2 import fakes as network_fakes
from openstackclient.tests.unit import utils as tests_utils


def _generate_fake_setting(attrs=None):
    setting_attrs = {
        'project_id': None,
        'stateful': False,
        'location': 'MUNCHMUNCHMUNCH',
    }
    if attrs:
        setting_attrs.update(attrs)
    return sdk_fakes.generate_fake_resource(
        _sg_default_statefulness.SecurityGroupsDefaultStatefulness,
        **setting_attrs,
    )


class TestCreateSecurityGroupDefaultStatefulness(
    network_fakes.TestNetworkV2,
):
    expected_columns = (
        'id',
        'project_id',
        'stateful',
    )

    def setUp(self):
        super().setUp()
        self.cmd = security_groups_default_statefulness.CreateSecurityGroupDefaultStatefulness(
            self.app, None
        )

    def test_create_stateless_system_wide(self):
        setting = _generate_fake_setting({'stateful': False})
        self.network_client.create_security_groups_default_statefulness.return_value = setting
        arglist = ['--stateless']
        verifylist = [('stateful', False)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_security_groups_default_statefulness.assert_called_once_with(
            **{'stateful': False, 'project_id': None}
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(
            (setting.id, setting.project_id, setting.stateful), data
        )

    def test_create_stateful_system_wide(self):
        setting = _generate_fake_setting({'stateful': True})
        self.network_client.create_security_groups_default_statefulness.return_value = setting
        arglist = ['--stateful']
        verifylist = [('stateful', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_security_groups_default_statefulness.assert_called_once_with(
            **{'stateful': True, 'project_id': None}
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(
            (setting.id, setting.project_id, setting.stateful), data
        )

    def test_create_per_project(self):
        project = sdk_fakes.generate_fake_resource(_project.Project)
        self.identity_sdk_client.find_project.return_value = project
        setting = _generate_fake_setting(
            {'stateful': False, 'project_id': project.id}
        )
        self.network_client.create_security_groups_default_statefulness.return_value = setting
        self.identity_client.projects.get.return_value = project
        arglist = ['--stateless', '--project', project.id]
        verifylist = [
            ('stateful', False),
            ('project', project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_security_groups_default_statefulness.assert_called_once_with(
            **{'stateful': False, 'project_id': project.id}
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(
            (setting.id, setting.project_id, setting.stateful), data
        )

    def test_create_no_statefulness_arg(self):
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            [],
            [],
        )


class TestDeleteSecurityGroupDefaultStatefulness(
    network_fakes.TestNetworkV2,
):
    def setUp(self):
        super().setUp()
        self.network_client.delete_security_groups_default_statefulness.return_value = None
        self.cmd = security_groups_default_statefulness.DeleteSecurityGroupDefaultStatefulness(
            self.app, None
        )
        self._settings = list(
            sdk_fakes.generate_fake_resources(
                _sg_default_statefulness.SecurityGroupsDefaultStatefulness,
                count=2,
                attrs={'stateful': False, 'location': 'MUNCHMUNCHMUNCH'},
            )
        )

    def test_delete_one(self):
        arglist = [self._settings[0].id]
        verifylist = [('setting', [self._settings[0].id])]
        self.network_client.find_security_groups_default_statefulness.return_value = self._settings[
            0
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.delete_security_groups_default_statefulness.assert_called_once_with(
            self._settings[0]
        )
        self.assertIsNone(result)

    def test_delete_multi(self):
        arglist = [s.id for s in self._settings]
        verifylist = [('setting', arglist)]
        self.network_client.find_security_groups_default_statefulness.side_effect = self._settings
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        calls = [call(s) for s in self._settings]
        self.network_client.delete_security_groups_default_statefulness.assert_has_calls(
            calls
        )
        self.assertIsNone(result)

    def test_delete_with_exception(self):
        arglist = [self._settings[0].id, 'nonexistent']
        verifylist = [
            ('setting', [self._settings[0].id, 'nonexistent']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        find_mock_result = [
            self._settings[0],
            exceptions.CommandError,
        ]
        self.network_client.find_security_groups_default_statefulness.side_effect = find_mock_result

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual(
                '1 of 2 default statefulness settings failed to delete.',
                str(e),
            )

        self.network_client.find_security_groups_default_statefulness.assert_any_call(
            self._settings[0].id, ignore_missing=False
        )
        self.network_client.find_security_groups_default_statefulness.assert_any_call(
            'nonexistent', ignore_missing=False
        )
        self.network_client.delete_security_groups_default_statefulness.assert_called_once_with(
            self._settings[0]
        )


class TestListSecurityGroupDefaultStatefulness(
    network_fakes.TestNetworkV2,
):
    def setUp(self):
        super().setUp()
        self._setting_system = _generate_fake_setting(
            {'stateful': False, 'project_id': None}
        )
        self._setting_project = _generate_fake_setting(
            {'stateful': True, 'project_id': 'project-id-1'}
        )
        self.expected_columns = (
            'ID',
            'Project ID',
            'Stateful',
        )
        self.expected_data = [
            (
                self._setting_system.id,
                self._setting_system.project_id,
                self._setting_system.stateful,
            ),
            (
                self._setting_project.id,
                self._setting_project.project_id,
                self._setting_project.stateful,
            ),
        ]
        self._settings = [self._setting_system, self._setting_project]
        self.network_client.security_groups_default_statefulness.return_value = self._settings
        self.cmd = security_groups_default_statefulness.ListSecurityGroupDefaultStatefulness(
            self.app, None
        )

    def test_list_all(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.security_groups_default_statefulness.assert_called_once_with()
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, list(data))


class TestSetSecurityGroupDefaultStatefulness(
    network_fakes.TestNetworkV2,
):
    def setUp(self):
        super().setUp()
        self._setting = _generate_fake_setting({'stateful': False})
        self.network_client.find_security_groups_default_statefulness.return_value = self._setting
        self.network_client.update_security_groups_default_statefulness.return_value = None
        self.cmd = security_groups_default_statefulness.SetSecurityGroupDefaultStatefulness(
            self.app, None
        )

    def test_set_stateful(self):
        arglist = [self._setting.id, '--stateful']
        verifylist = [
            ('setting', self._setting.id),
            ('stateful', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network_client.update_security_groups_default_statefulness.assert_called_once_with(
            self._setting, **{'stateful': True}
        )
        self.assertIsNone(result)

    def test_set_stateless(self):
        arglist = [self._setting.id, '--stateless']
        verifylist = [
            ('setting', self._setting.id),
            ('stateful', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network_client.update_security_groups_default_statefulness.assert_called_once_with(
            self._setting, **{'stateful': False}
        )
        self.assertIsNone(result)

    def test_set_no_statefulness_arg(self):
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            [self._setting.id],
            [],
        )


class TestShowSecurityGroupDefaultStatefulness(
    network_fakes.TestNetworkV2,
):
    expected_columns = (
        'id',
        'project_id',
        'stateful',
    )

    def setUp(self):
        super().setUp()
        self._setting = _generate_fake_setting()
        self.network_client.find_security_groups_default_statefulness.return_value = self._setting
        self.cmd = security_groups_default_statefulness.ShowSecurityGroupDefaultStatefulness(
            self.app, None
        )

    def test_show(self):
        arglist = [self._setting.id]
        verifylist = [('setting', self._setting.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.find_security_groups_default_statefulness.assert_called_once_with(
            self._setting.id, ignore_missing=False
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(
            (
                self._setting.id,
                self._setting.project_id,
                self._setting.stateful,
            ),
            data,
        )



#   Copyright 2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""Agent action implementations"""

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from openstack import exceptions as sdk_exceptions
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


LOG = logging.getLogger(__name__)


class CreateAgent(command.ShowOne):
    """Create compute agent.

    The compute agent functionality is hypervisor specific and is only
    supported by the XenAPI hypervisor driver. It was removed from nova in the
    23.0.0 (Wallaby) release.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument("os", metavar="<os>", help=_("Type of OS"))
        parser.add_argument(
            "architecture",
            metavar="<architecture>",
            help=_("Type of architecture"),
        )
        parser.add_argument("version", metavar="<version>", help=_("Version"))
        parser.add_argument("url", metavar="<url>", help=_("URL"))
        parser.add_argument("md5hash", metavar="<md5hash>", help=_("MD5 hash"))
        parser.add_argument(
            "hypervisor",
            metavar="<hypervisor>",
            default="xen",
            help=_("Type of hypervisor"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        compute_client = self.app.client_manager.compute

        # doing this since openstacksdk has decided not to support this
        # deprecated command
        data = {
            'agent': {
                'hypervisor': parsed_args.hypervisor,
                'os': parsed_args.os,
                'architecture': parsed_args.architecture,
                'version': parsed_args.version,
                'url': parsed_args.url,
                'md5hash': parsed_args.md5hash,
            },
        }
        response = compute_client.post(
            '/os-agents', json=data, microversion='2.1'
        )
        sdk_exceptions.raise_from_response(response)
        agent = response.json().get('agent')

        col_headers, col_data = zip(*sorted(agent.items()))
        return col_headers, col_data


class DeleteAgent(command.Command):
    """Delete compute agent(s).

    The compute agent functionality is hypervisor specific and is only
    supported by the XenAPI hypervisor driver. It was removed from nova in the
    23.0.0 (Wallaby) release.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "id", metavar="<id>", nargs='+', help=_("ID of agent(s) to delete")
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        compute_client = self.app.client_manager.compute
        result = 0
        for id in parsed_args.id:
            try:
                # doing this since openstacksdk has decided not to support this
                # deprecated command
                response = compute_client.delete(
                    f'/os-agents/{id}', microversion='2.1'
                )
                sdk_exceptions.raise_from_response(response)
            except Exception as e:
                result += 1
                LOG.error(
                    _("Failed to delete agent with ID '%(id)s': %(e)s"),
                    {'id': id, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.id)
            msg = _("%(result)s of %(total)s agents failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListAgent(command.Lister):
    """List compute agents.

    The compute agent functionality is hypervisor specific and is only
    supported by the XenAPI hypervisor driver. It was removed from nova in the
    23.0.0 (Wallaby) release.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--hypervisor",
            metavar="<hypervisor>",
            help=_("Type of hypervisor"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        compute_client = self.app.client_manager.compute
        columns = (
            "Agent ID",
            "Hypervisor",
            "OS",
            "Architecture",
            "Version",
            "Md5Hash",
            "URL",
        )

        # doing this since openstacksdk has decided not to support this
        # deprecated command
        path = '/os-agents'
        if parsed_args.hypervisor:
            path += f'?hypervisor={parsed_args.hypervisor}'

        response = compute_client.get(path, microversion='2.1')
        sdk_exceptions.raise_from_response(response)
        agents = response.json().get('agents')

        return columns, (utils.get_dict_properties(s, columns) for s in agents)


class SetAgent(command.Command):
    """Set compute agent properties.

    The compute agent functionality is hypervisor-specific and is only
    supported by the XenAPI hypervisor driver. It was removed from nova in the
    23.0.0 (Wallaby) release.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "id",
            metavar="<id>",
            type=int,
            help=_("ID of the agent"),
        )
        parser.add_argument(
            "--agent-version",
            dest="version",
            metavar="<version>",
            help=_("Version of the agent"),
        )
        parser.add_argument(
            "--url", metavar="<url>", help=_("URL of the agent")
        )
        parser.add_argument(
            "--md5hash", metavar="<md5hash>", help=_("MD5 hash of the agent")
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        compute_client = self.app.client_manager.compute

        response = compute_client.get('/os-agents', microversion='2.1')
        sdk_exceptions.raise_from_response(response)
        agents = response.json().get('agents')
        data = {}
        for agent in agents:
            if agent['agent_id'] == parsed_args.id:
                data['version'] = agent['version']
                data['url'] = agent['url']
                data['md5hash'] = agent['md5hash']
                break
        else:
            msg = _("No agent with a ID of '%(id)s' exists.")
            raise exceptions.CommandError(msg % {'id': parsed_args.id})

        if parsed_args.version:
            data['version'] = parsed_args.version
        if parsed_args.url:
            data['url'] = parsed_args.url
        if parsed_args.md5hash:
            data['md5hash'] = parsed_args.md5hash

        data = {'para': data}

        # doing this since openstacksdk has decided not to support this
        # deprecated command
        response = compute_client.put(
            f'/os-agents/{parsed_args.id}', json=data, microversion='2.1'
        )
        sdk_exceptions.raise_from_response(response)



#   Copyright 2016 NEC Corporation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest import mock

from osc_lib import exceptions

from openstackclient import command
from openstackclient.tests.unit import fakes as test_fakes
from openstackclient.tests.unit import utils as test_utils


class FakeCommand(command.Command):
    def take_action(self, parsed_args):
        pass


class TestCommand(test_utils.TestCase):
    def test_command_has_logger(self):
        cmd = FakeCommand(mock.Mock(), mock.Mock())
        self.assertTrue(hasattr(cmd, 'log'))
        self.assertEqual(
            'openstackclient.tests.unit.common.test_command.FakeCommand',
            cmd.log.name,
        )

    def test_validate_os_beta_command_enabled(self):
        cmd = FakeCommand(mock.Mock(), mock.Mock())
        cmd.app = mock.Mock()
        cmd.app.options = test_fakes.FakeOptions()

        # No exception is raised when enabled.
        cmd.app.options.os_beta_command = True
        cmd.validate_os_beta_command_enabled()

        cmd.app.options.os_beta_command = False
        self.assertRaises(
            exceptions.CommandError, cmd.validate_os_beta_command_enabled
        )



#   Copyright 2019 SUSE LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest.mock import call

from openstack import exceptions as sdk_exceptions
from openstack.identity.v3 import access_rule as _access_rule
from openstack.test import fakes as sdk_fakes
from osc_lib import exceptions

from openstackclient.identity.v3 import access_rule
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes


class TestAccessRuleDelete(identity_fakes.TestIdentity):
    access_rule = sdk_fakes.generate_fake_resource(_access_rule.AccessRule)

    def setUp(self):
        super().setUp()

        self.identity_sdk_client.get_access_rule.return_value = (
            self.access_rule
        )
        self.identity_sdk_client.delete_access_rule.return_value = None

        # Get the command object to test
        self.cmd = access_rule.DeleteAccessRule(self.app, None)

    def test_access_rule_delete(self):
        arglist = [self.access_rule.id]
        verifylist = [('access_rule', [self.access_rule.id])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        conn = self.app.client_manager.sdk_connection
        user_id = conn.config.get_auth().get_user_id(conn.identity)

        result = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.delete_access_rule.assert_called_with(
            user_id,
            self.access_rule.id,
        )
        self.assertIsNone(result)

    def test_delete_multi_access_rules_with_exception(self):
        self.identity_sdk_client.get_access_rule.side_effect = [
            self.access_rule,
            sdk_exceptions.NotFoundException,
        ]

        arglist = [
            self.access_rule.id,
            'nonexistent_access_rule',
        ]
        verifylist = [
            ('access_rule', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        conn = self.app.client_manager.sdk_connection
        user_id = conn.config.get_auth().get_user_id(conn.identity)

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 access rules failed to delete.', str(e))

        calls = []
        for a in arglist:
            calls.append(call(user_id, a))

        self.identity_sdk_client.get_access_rule.assert_has_calls(calls)

        self.assertEqual(
            2, self.identity_sdk_client.get_access_rule.call_count
        )
        self.identity_sdk_client.delete_access_rule.assert_called_once_with(
            user_id, self.access_rule.id
        )


class TestAccessRuleList(identity_fakes.TestIdentity):
    access_rule = sdk_fakes.generate_fake_resource(_access_rule.AccessRule)

    def setUp(self):
        super().setUp()

        self.identity_sdk_client.access_rules.return_value = [self.access_rule]

        # Get the command object to test
        self.cmd = access_rule.ListAccessRule(self.app, None)

    def test_access_rule_list(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        conn = self.app.client_manager.sdk_connection
        user_id = conn.config.get_auth().get_user_id(conn.identity)

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.access_rules.assert_called_with(user=user_id)

        collist = ('ID', 'Service', 'Method', 'Path')
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.access_rule.id,
                self.access_rule.service,
                self.access_rule.method,
                self.access_rule.path,
            ),
        )
        self.assertEqual(datalist, tuple(data))


class TestAccessRuleShow(identity_fakes.TestIdentity):
    access_rule = sdk_fakes.generate_fake_resource(_access_rule.AccessRule)

    def setUp(self):
        super().setUp()

        self.identity_sdk_client.get_access_rule.return_value = (
            self.access_rule
        )

        # Get the command object to test
        self.cmd = access_rule.ShowAccessRule(self.app, None)

    def test_access_rule_show(self):
        arglist = [
            self.access_rule.id,
        ]
        verifylist = [
            ('access_rule', self.access_rule.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        conn = self.app.client_manager.sdk_connection
        user_id = conn.config.get_auth().get_user_id(conn.identity)

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.get_access_rule.assert_called_with(
            user_id, self.access_rule.id
        )

        collist = ('ID', 'Method', 'Path', 'Service')
        self.assertEqual(collist, columns)
        datalist = (
            self.access_rule.id,
            self.access_rule.method,
            self.access_rule.path,
            self.access_rule.service,
        )
        self.assertEqual(datalist, data)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib.common.utils import data_utils
from tempest.lib import exceptions

from openstackclient.tests.functional.identity.v2 import common


class UserTests(common.IdentityTests):
    def test_user_create(self):
        self._create_dummy_user()

    def test_user_delete(self):
        username = self._create_dummy_user(add_clean_up=False)
        raw_output = self.openstack(f'user delete {username}')
        self.assertEqual(0, len(raw_output))

    def test_user_list(self):
        raw_output = self.openstack('user list')
        items = self.parse_listing(raw_output)
        self.assert_table_structure(items, common.BASIC_LIST_HEADERS)

    def test_user_set(self):
        username = self._create_dummy_user()
        raw_output = self.openstack(f'user show {username}')
        user = self.parse_show_as_object(raw_output)
        new_username = data_utils.rand_name('NewTestUser')
        new_email = data_utils.rand_name() + '@example.com'
        raw_output = self.openstack(
            'user set --email {email} --name {new_name} {id}'.format(
                email=new_email, new_name=new_username, id=user['id']
            )
        )
        self.assertEqual(0, len(raw_output))
        raw_output = self.openstack(f'user show {new_username}')
        new_user = self.parse_show_as_object(raw_output)
        self.assertEqual(user['id'], new_user['id'])
        self.assertEqual(new_email, new_user['email'])

    def test_user_show(self):
        username = self._create_dummy_user()
        raw_output = self.openstack(f'user show {username}')
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.USER_FIELDS)

    def test_bad_user_command(self):
        self.assertRaises(
            exceptions.CommandFailed, self.openstack, 'user unlist'
        )



#    Copyright 2018 SUSE Linux GmbH
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from tempest.lib.common.utils import data_utils

from openstackclient.tests.functional.identity.v3 import common


class ApplicationCredentialTests(common.IdentityTests):
    APPLICATION_CREDENTIAL_FIELDS = [
        'ID',
        'Name',
        'Project ID',
        'Description',
        'Roles',
        'Expires At',
        'Unrestricted',
    ]
    APPLICATION_CREDENTIAL_LIST_HEADERS = [
        'ID',
        'Name',
        'Project ID',
        'Description',
        'Expires At',
    ]

    def test_application_credential_create(self):
        name = data_utils.rand_name('name')
        raw_output = self.openstack(f'application credential create {name}')
        self.addCleanup(
            self.openstack,
            f'application credential delete {name}',
        )
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.APPLICATION_CREDENTIAL_FIELDS)

    def _create_role_assignments(self):
        try:
            user = self.openstack(
                'configuration show -f value -c auth.username'
            )
        except Exception:
            user = self.openstack(
                'configuration show -f value -c auth.user_id'
            )
        try:
            user_domain = self.openstack(
                'configuration show -f value -c auth.user_domain_name'
            )
        except Exception:
            user_domain = self.openstack(
                'configuration show -f value -c auth.user_domain_id'
            )
        try:
            project = self.openstack(
                'configuration show -f value -c auth.project_name'
            )
        except Exception:
            project = self.openstack(
                'configuration show -f value -c auth.project_id'
            )
        try:
            project_domain = self.openstack(
                'configuration show -f value -c auth.project_domain_name'
            )
        except Exception:
            project_domain = self.openstack(
                'configuration show -f value -c auth.project_domain_id'
            )
        role1 = self._create_dummy_role()
        role2 = self._create_dummy_role()
        for role in role1, role2:
            self.openstack(
                'role add'
                f' --user {user}'
                f' --user-domain {user_domain}'
                f' --project {project}'
                f' --project-domain {project_domain}'
                f' {role}'
            )
            self.addCleanup(
                self.openstack,
                'role remove'
                f' --user {user}'
                f' --user-domain {user_domain}'
                f' --project {project}'
                f' --project-domain {project_domain}'
                f' {role}',
            )
        return role1, role2

    def test_application_credential_create_with_options(self):
        name = data_utils.rand_name('name')
        secret = data_utils.rand_name('secret')
        description = data_utils.rand_name('description')
        tomorrow = (
            datetime.datetime.now(datetime.UTC).replace(tzinfo=None)
            + datetime.timedelta(days=1)
        ).strftime('%Y-%m-%dT%H:%M:%S%z')
        role1, role2 = self._create_role_assignments()
        raw_output = self.openstack(
            f'application credential create {name}'
            f' --secret {secret}'
            f' --description {description}'
            f' --expiration {tomorrow}'
            f' --role {role1}'
            f' --role {role2}'
            ' --unrestricted'
        )
        self.addCleanup(
            self.openstack,
            f'application credential delete {name}',
        )
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.APPLICATION_CREDENTIAL_FIELDS)

    def test_application_credential_delete(self):
        name = data_utils.rand_name('name')
        self.openstack(f'application credential create {name}')
        raw_output = self.openstack(f'application credential delete {name}')
        self.assertEqual(0, len(raw_output))

    def test_application_credential_list(self):
        raw_output = self.openstack('application credential list')
        items = self.parse_listing(raw_output)
        self.assert_table_structure(
            items, self.APPLICATION_CREDENTIAL_LIST_HEADERS
        )

    def test_application_credential_show(self):
        name = data_utils.rand_name('name')
        raw_output = self.openstack(f'application credential create {name}')
        self.addCleanup(
            self.openstack,
            f'application credential delete {name}',
        )
        raw_output = self.openstack(f'application credential show {name}')
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.APPLICATION_CREDENTIAL_FIELDS)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
from collections.abc import Iterable, Sequence
from typing import Any

from osc_lib.cli import identity as identity_utils
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common
from openstackclient.network.v2.dynamic_routing import bgp_peer


def _get_attrs(parsed_args: argparse.Namespace) -> dict[str, Any]:
    attrs: dict[str, Any] = {}
    if parsed_args.name is not None:
        attrs['name'] = str(parsed_args.name)
    if 'local_as' in parsed_args:
        attrs['local_as'] = parsed_args.local_as
    if 'ip_version' in parsed_args:
        attrs['ip_version'] = parsed_args.ip_version
    if parsed_args.advertise_tenant_networks:
        attrs['advertise_tenant_networks'] = True
    if parsed_args.no_advertise_tenant_networks:
        attrs['advertise_tenant_networks'] = False
    if parsed_args.advertise_floating_ip_host_routes:
        attrs['advertise_floating_ip_host_routes'] = True
    if parsed_args.no_advertise_floating_ip_host_routes:
        attrs['advertise_floating_ip_host_routes'] = False
    return attrs


def add_common_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        '--advertise-floating-ip-host-routes',
        action='store_true',
        help=_(
            "Enable the advertisement of floating IP host routes "
            "by the BGP speaker. (default)"
        ),
    )
    parser.add_argument(
        '--no-advertise-floating-ip-host-routes',
        action='store_true',
        help=_(
            "Disable the advertisement of floating IP host routes "
            "by the BGP speaker."
        ),
    )
    parser.add_argument(
        '--advertise-tenant-networks',
        action='store_true',
        help=_(
            "Enable the advertisement of tenant network routes "
            "by the BGP speaker. (default)"
        ),
    )
    parser.add_argument(
        '--no-advertise-tenant-networks',
        action='store_true',
        help=_(
            "Disable the advertisement of tenant network routes "
            "by the BGP speaker."
        ),
    )


class AddNetworkToSpeaker(command.Command):
    _description = _("Add a network to a BGP speaker")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'bgp_speaker',
            metavar='<bgp-speaker>',
            help=_("BGP speaker (name or ID)"),
        )
        parser.add_argument(
            'network',
            metavar='<network>',
            help=_("Network to add (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        speaker_id = client.find_bgp_speaker(
            parsed_args.bgp_speaker, ignore_missing=False
        ).id
        net_id = client.find_network(
            parsed_args.network, ignore_missing=False
        ).id
        client.add_gateway_network_to_speaker(speaker_id, net_id)


class AddPeerToSpeaker(command.Command):
    _description = _("Add a peer to a BGP speaker")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'bgp_speaker',
            metavar='<bgp-speaker>',
            help=_("BGP speaker (name or ID)"),
        )
        parser.add_argument(
            'bgp_peer',
            metavar='<bgp-peer>',
            help=_("BGP Peer to add (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        speaker_id = client.find_bgp_speaker(
            parsed_args.bgp_speaker, ignore_missing=False
        )['id']
        peer_id = client.find_bgp_peer(
            parsed_args.bgp_peer, ignore_missing=False
        )['id']
        client.add_bgp_peer_to_speaker(speaker_id, peer_id)


class CreateBgpSpeaker(command.ShowOne):
    _description = _("Create a BGP speaker")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help=_("Name of the BGP speaker to create"),
        )
        parser.add_argument(
            '--local-as',
            metavar='<local-as>',
            required=True,
            help=_(
                "Local AS number. (Integer in [%(min_val)s, %(max_val)s] "
                "is allowed.)"
            )
            % {
                'min_val': bgp_peer.MIN_AS_NUM,
                'max_val': bgp_peer.MAX_AS_NUM,
            },
        )
        parser.add_argument(
            '--ip-version',
            type=int,
            choices=[4, 6],
            default=4,
            help=_("IP version for the BGP speaker (default is 4)"),
        )
        add_common_arguments(parser)
        identity_utils.add_project_owner_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network

        attrs = _get_attrs(parsed_args)
        if parsed_args.project is not None:
            identity_client = self.app.client_manager.sdk_connection.identity
            project_id = identity_common.find_project_id_sdk(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            )
            attrs['project_id'] = project_id

        obj = client.create_bgp_speaker(**attrs)
        display_columns, columns = utils.get_osc_show_columns_for_sdk_resource(
            obj, {}, ['location', 'tenant_id']
        )
        data = utils.get_dict_properties(obj, columns)
        return display_columns, data


class DeleteBgpSpeaker(command.Command):
    _description = _("Delete a BGP speaker")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'bgp_speaker',
            metavar="<bgp-speaker>",
            help=_("BGP speaker to delete (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        id = client.find_bgp_speaker(
            parsed_args.bgp_speaker, ignore_missing=False
        ).id
        client.delete_bgp_speaker(id)


class ListBgpSpeaker(command.Lister):
    _description = _("List BGP speakers")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--agent',
            metavar='<agent-id>',
            help=_("List BGP speakers hosted by an agent (ID only)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        data: Iterable[Any]
        if parsed_args.agent is not None:
            data = client.get_bgp_speakers_hosted_by_dragent(parsed_args.agent)
        else:
            data = client.bgp_speakers(retrieve_all=True)

        headers = ('ID', 'Name', 'Local AS', 'IP Version')
        columns = ('id', 'name', 'local_as', 'ip_version')
        return (headers, (utils.get_dict_properties(s, columns) for s in data))


class ListRoutesAdvertisedBySpeaker(command.Lister):
    _description = _("List routes advertised")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'bgp_speaker',
            metavar='<bgp-speaker>',
            help=_("BGP speaker (name or ID)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        speaker_id = client.find_bgp_speaker(
            parsed_args.bgp_speaker, ignore_missing=False
        )['id']
        data = client.get_advertised_routes_of_speaker(speaker_id)
        headers = ('Destination', 'Nexthop')
        columns = ('destination', 'next_hop')
        return (
            headers,
            (
                utils.get_dict_properties(s, columns)
                for s in data['advertised_routes']
            ),
        )


class RemoveNetworkFromSpeaker(command.Command):
    _description = _("Remove a network from a BGP speaker")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'bgp_speaker',
            metavar='<bgp-speaker>',
            help=_("BGP speaker (name or ID)"),
        )
        parser.add_argument(
            'network',
            metavar='<network>',
            help=_("Network to remove (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        speaker_id = client.find_bgp_speaker(
            parsed_args.bgp_speaker, ignore_missing=False
        )['id']
        net_id = client.find_network(
            parsed_args.network, ignore_missing=False
        )['id']
        client.remove_gateway_network_from_speaker(speaker_id, net_id)


class RemovePeerFromSpeaker(command.Command):
    _description = _("Remove a peer from a BGP speaker")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'bgp_speaker',
            metavar='<bgp-speaker>',
            help=_("BGP speaker (name or ID)"),
        )
        parser.add_argument(
            'bgp_peer',
            metavar='<bgp-peer>',
            help=_("BGP Peer to remove (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        speaker_id = client.find_bgp_speaker(
            parsed_args.bgp_speaker, ignore_missing=False
        )['id']
        peer_id = client.find_bgp_peer(
            parsed_args.bgp_peer, ignore_missing=False
        )['id']
        client.remove_bgp_peer_from_speaker(speaker_id, peer_id)


class SetBgpSpeaker(command.Command):
    _description = _("Set BGP speaker properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'bgp_speaker',
            metavar="<bgp-speaker>",
            help=_("BGP speaker to update (name or ID)"),
        )
        parser.add_argument('--name', help=_("New name for the BGP speaker"))
        add_common_arguments(parser)
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        id = client.find_bgp_speaker(
            parsed_args.bgp_speaker, ignore_missing=False
        ).id
        attrs = _get_attrs(parsed_args)
        client.update_bgp_speaker(id, **attrs)


class ShowBgpSpeaker(command.ShowOne):
    _description = _("Show a BGP speaker")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'bgp_speaker',
            metavar="<bgp-speaker>",
            help=_("BGP speaker to display (name or ID)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        obj = client.find_bgp_speaker(
            parsed_args.bgp_speaker, ignore_missing=False
        )
        display_columns, columns = utils.get_osc_show_columns_for_sdk_resource(
            obj, {}, ['location', 'tenant_id']
        )
        data = utils.get_dict_properties(obj, columns)
        return display_columns, data



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest import mock

from openstack.block_storage.v3 import backup as _backup
from openstack.block_storage.v3 import snapshot as _snapshot
from openstack.block_storage.v3 import volume as _volume
from openstack import exceptions as sdk_exceptions
from openstack.identity.v3 import project as _project
from openstack.test import fakes as sdk_fakes
from osc_lib import exceptions

from openstackclient.tests.unit.volume.v3 import fakes as volume_fakes
from openstackclient.volume.v3 import volume_backup


class TestBackupCreate(volume_fakes.TestVolume):
    columns = (
        'id',
        'name',
        'volume_id',
    )

    def setUp(self):
        super().setUp()

        self.volume = sdk_fakes.generate_fake_resource(_volume.Volume)
        self.volume_client.find_volume.return_value = self.volume
        self.snapshot = sdk_fakes.generate_fake_resource(_snapshot.Snapshot)
        self.volume_client.find_snapshot.return_value = self.snapshot
        self.backup = sdk_fakes.generate_fake_resource(
            _backup.Backup,
            volume_id=self.volume.id,
            snapshot_id=self.snapshot.id,
        )
        self.volume_client.create_backup.return_value = self.backup

        self.data = (
            self.backup.id,
            self.backup.name,
            self.backup.volume_id,
        )

        self.cmd = volume_backup.CreateVolumeBackup(self.app, None)

    def test_backup_create(self):
        arglist = [
            "--name",
            self.backup.name,
            "--description",
            self.backup.description,
            "--container",
            self.backup.container,
            "--force",
            "--incremental",
            "--snapshot",
            self.backup.snapshot_id,
            self.backup.volume_id,
        ]
        verifylist = [
            ("name", self.backup.name),
            ("description", self.backup.description),
            ("container", self.backup.container),
            ("force", True),
            ("incremental", True),
            ("snapshot", self.backup.snapshot_id),
            ("volume", self.backup.volume_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.create_backup.assert_called_with(
            volume_id=self.backup.volume_id,
            container=self.backup.container,
            name=self.backup.name,
            description=self.backup.description,
            force=True,
            is_incremental=True,
            snapshot_id=self.backup.snapshot_id,
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_backup_create_with_properties(self):
        self.set_volume_api_version('3.43')

        arglist = [
            "--property",
            "foo=bar",
            "--property",
            "wow=much-cool",
            self.backup.volume_id,
        ]
        verifylist = [
            ("properties", {"foo": "bar", "wow": "much-cool"}),
            ("volume", self.backup.volume_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.create_backup.assert_called_with(
            volume_id=self.backup.volume_id,
            container=None,
            name=None,
            description=None,
            force=False,
            is_incremental=False,
            metadata={"foo": "bar", "wow": "much-cool"},
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_backup_create_with_properties_pre_v343(self):
        self.set_volume_api_version('3.42')

        arglist = [
            "--property",
            "foo=bar",
            "--property",
            "wow=much-cool",
            self.backup.volume_id,
        ]
        verifylist = [
            ("properties", {"foo": "bar", "wow": "much-cool"}),
            ("volume", self.backup.volume_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn("--os-volume-api-version 3.43 or greater", str(exc))

    def test_backup_create_with_availability_zone(self):
        self.set_volume_api_version('3.51')

        arglist = [
            "--availability-zone",
            "my-az",
            self.backup.volume_id,
        ]
        verifylist = [
            ("availability_zone", "my-az"),
            ("volume", self.backup.volume_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.create_backup.assert_called_with(
            volume_id=self.backup.volume_id,
            container=None,
            name=None,
            description=None,
            force=False,
            is_incremental=False,
            availability_zone="my-az",
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_backup_create_with_availability_zone_pre_v351(self):
        self.set_volume_api_version('3.50')

        arglist = [
            "--availability-zone",
            "my-az",
            self.backup.volume_id,
        ]
        verifylist = [
            ("availability_zone", "my-az"),
            ("volume", self.backup.volume_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn("--os-volume-api-version 3.51 or greater", str(exc))

    def test_backup_create_without_name(self):
        arglist = [
            "--description",
            self.backup.description,
            "--container",
            self.backup.container,
            self.backup.volume_id,
        ]
        verifylist = [
            ("description", self.backup.description),
            ("container", self.backup.container),
            ("volume", self.backup.volume_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.create_backup.assert_called_with(
            volume_id=self.backup.volume_id,
            container=self.backup.container,
            name=None,
            description=self.backup.description,
            force=False,
            is_incremental=False,
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)


class TestBackupDelete(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.backups = list(sdk_fakes.generate_fake_resources(_backup.Backup))
        self.volume_client.find_backup.side_effect = self.backups
        self.volume_client.delete_backup.return_value = None

        self.cmd = volume_backup.DeleteVolumeBackup(self.app, None)

    def test_backup_delete(self):
        arglist = [self.backups[0].id]
        verifylist = [("backups", [self.backups[0].id])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.delete_backup.assert_called_with(
            self.backups[0].id, ignore_missing=False, force=False
        )
        self.assertIsNone(result)

    def test_backup_delete_with_force(self):
        arglist = [
            '--force',
            self.backups[0].id,
        ]
        verifylist = [('force', True), ("backups", [self.backups[0].id])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.delete_backup.assert_called_with(
            self.backups[0].id, ignore_missing=False, force=True
        )
        self.assertIsNone(result)

    def test_delete_multiple_backups(self):
        arglist = []
        for b in self.backups:
            arglist.append(b.id)
        verifylist = [
            ('backups', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        calls = []
        for b in self.backups:
            calls.append(mock.call(b.id, ignore_missing=False, force=False))
        self.volume_client.delete_backup.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_delete_multiple_backups_with_exception(self):
        arglist = [
            self.backups[0].id,
            'unexist_backup',
        ]
        verifylist = [
            ('backups', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        find_mock_result = [self.backups[0], exceptions.CommandError]
        self.volume_client.find_backup.side_effect = find_mock_result

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 backups failed to delete.', str(e))

        self.volume_client.find_backup.assert_any_call(
            self.backups[0].id, ignore_missing=False
        )
        self.volume_client.find_backup.assert_any_call(
            'unexist_backup', ignore_missing=False
        )

        self.assertEqual(2, self.volume_client.find_backup.call_count)
        self.volume_client.delete_backup.assert_called_once_with(
            self.backups[0].id,
            ignore_missing=False,
            force=False,
        )


class TestBackupList(volume_fakes.TestVolume):
    columns = (
        'ID',
        'Name',
        'Description',
        'Status',
        'Size',
        'Incremental',
        'Created At',
    )
    columns_long = (*columns, 'Availability Zone', 'Volume', 'Container')

    def setUp(self):
        super().setUp()

        self.volume = sdk_fakes.generate_fake_resource(_volume.Volume)
        self.volume_client.find_volume.return_value = self.volume
        self.volume_client.volumes.return_value = [self.volume]
        self.backups = list(
            sdk_fakes.generate_fake_resources(
                _backup.Backup,
                attrs={'volume_id': self.volume.id},
            )
        )
        self.volume_client.backups.return_value = self.backups
        self.volume_client.find_backup.return_value = self.backups[0]

        self.data = []
        for b in self.backups:
            self.data.append(
                (
                    b.id,
                    b.name,
                    b.description,
                    b.status,
                    b.size,
                    b.is_incremental,
                    b.created_at,
                )
            )
        self.data_long = []
        for b in self.backups:
            self.data_long.append(
                (
                    b.id,
                    b.name,
                    b.description,
                    b.status,
                    b.size,
                    b.is_incremental,
                    b.created_at,
                    b.availability_zone,
                    volume_backup.VolumeIdColumn(b.volume_id),
                    b.container,
                )
            )

        self.cmd = volume_backup.ListVolumeBackup(self.app, None)

    def test_backup_list_without_options(self):
        arglist = []
        verifylist = [
            ("long", False),
            ("name", None),
            ("status", None),
            ("volume", None),
            ("marker", None),
            ("limit", None),
            ('all_projects', False),
            ("project", None),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.find_volume.assert_not_called()
        self.volume_client.find_backup.assert_not_called()
        self.volume_client.backups.assert_called_with(
            name=None,
            status=None,
            volume_id=None,
            all_tenants=False,
            marker=None,
            limit=None,
            max_items=None,
            project_id=None,
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_backup_list_with_options(self):
        project = sdk_fakes.generate_fake_resource(_project.Project)
        self.identity_sdk_client.find_project.return_value = project
        arglist = [
            "--long",
            "--name",
            self.backups[0].name,
            "--status",
            "error",
            "--volume",
            self.volume.id,
            "--marker",
            self.backups[0].id,
            "--all-projects",
            "--limit",
            "3",
            "--project",
            project.id,
        ]
        verifylist = [
            ("long", True),
            ("name", self.backups[0].name),
            ("status", "error"),
            ("volume", self.volume.id),
            ("marker", self.backups[0].id),
            ('all_projects', True),
            ("limit", 3),
            ("project", project.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.find_volume.assert_called_once_with(
            self.volume.id, ignore_missing=False
        )
        self.volume_client.find_backup.assert_called_once_with(
            self.backups[0].id, ignore_missing=False
        )
        self.volume_client.backups.assert_called_with(
            name=self.backups[0].name,
            status="error",
            volume_id=self.volume.id,
            all_tenants=True,
            marker=self.backups[0].id,
            limit=3,
            max_items=None,
            project_id=project.id,
        )
        self.assertEqual(self.columns_long, columns)
        self.assertCountEqual(self.data_long, list(data))


class TestBackupRestore(volume_fakes.TestVolume):
    columns = (
        "id",
        "volume_id",
        "volume_name",
    )

    def setUp(self):
        super().setUp()

        self.volume = sdk_fakes.generate_fake_resource(_volume.Volume)
        self.volume_client.find_volume.return_value = self.volume
        self.backup = sdk_fakes.generate_fake_resource(
            _backup.Backup, volume_id=self.volume.id
        )
        self.volume_client.find_backup.return_value = self.backup
        self.volume_client.create_backup.return_value = self.backup
        self.volume_client.restore_backup.return_value = {
            'id': self.backup['id'],
            'volume_id': self.volume['id'],
            'volume_name': self.volume['name'],
        }

        self.data = (
            self.backup.id,
            self.volume.id,
            self.volume.name,
        )

        self.cmd = volume_backup.RestoreVolumeBackup(self.app, None)

    def test_backup_restore(self):
        self.volume_client.find_volume.side_effect = exceptions.CommandError()
        arglist = [self.backup.id]
        verifylist = [
            ("backup", self.backup.id),
            ("volume", None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_client.restore_backup.assert_called_with(
            self.backup.id, volume=None, name=None
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_backup_restore_with_volume(self):
        self.volume_client.find_volume.side_effect = exceptions.CommandError()
        arglist = [
            self.backup.id,
            self.backup.volume_id,
        ]
        verifylist = [
            ("backup", self.backup.id),
            ("volume", self.backup.volume_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_client.restore_backup.assert_called_with(
            self.backup.id, volume=None, name=self.backup.volume_id
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_backup_restore_with_volume_force(self):
        arglist = [
            "--force",
            self.backup.id,
            self.volume.name,
        ]
        verifylist = [
            ("force", True),
            ("backup", self.backup.id),
            ("volume", self.volume.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_client.restore_backup.assert_called_with(
            self.backup.id, volume=self.volume.id, name=None
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_backup_restore_with_volume_existing(self):
        arglist = [
            self.backup.id,
            self.volume.name,
        ]
        verifylist = [
            ("backup", self.backup.id),
            ("volume", self.volume.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )


class TestBackupSet(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.backup = sdk_fakes.generate_fake_resource(
            _backup.Backup, metadata={'wow': 'cool'}
        )
        self.volume_client.find_backup.return_value = self.backup

        self.cmd = volume_backup.SetVolumeBackup(self.app, None)

    def test_backup_set_name(self):
        self.set_volume_api_version('3.9')

        arglist = [
            '--name',
            'new_name',
            self.backup.id,
        ]
        verifylist = [
            ('name', 'new_name'),
            ('backup', self.backup.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.volume_client.find_backup.assert_called_with(
            self.backup.id, ignore_missing=False
        )
        self.volume_client.update_backup.assert_called_once_with(
            self.backup, name='new_name'
        )

    def test_backup_set_name_pre_v39(self):
        self.set_volume_api_version('3.8')

        arglist = [
            '--name',
            'new_name',
            self.backup.id,
        ]
        verifylist = [
            ('name', 'new_name'),
            ('backup', self.backup.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn("--os-volume-api-version 3.9 or greater", str(exc))

    def test_backup_set_description(self):
        self.set_volume_api_version('3.9')

        arglist = [
            '--description',
            'new_description',
            self.backup.id,
        ]
        verifylist = [
            ('name', None),
            ('description', 'new_description'),
            ('backup', self.backup.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.volume_client.find_backup.assert_called_with(
            self.backup.id, ignore_missing=False
        )
        self.volume_client.update_backup.assert_called_once_with(
            self.backup, description='new_description'
        )

    def test_backup_set_description_pre_v39(self):
        self.set_volume_api_version('3.8')

        arglist = [
            '--description',
            'new_description',
            self.backup.id,
        ]
        verifylist = [
            ('name', None),
            ('description', 'new_description'),
            ('backup', self.backup.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn("--os-volume-api-version 3.9 or greater", str(exc))

    def test_backup_set_state(self):
        arglist = ['--state', 'error', self.backup.id]
        verifylist = [('state', 'error'), ('backup', self.backup.id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.volume_client.find_backup.assert_called_with(
            self.backup.id, ignore_missing=False
        )
        self.volume_client.reset_backup_status.assert_called_with(
            self.backup, status='error'
        )

    def test_backup_set_state_failed(self):
        self.volume_client.reset_backup_status.side_effect = (
            sdk_exceptions.NotFoundException('foo')
        )

        arglist = ['--state', 'error', self.backup.id]
        verifylist = [('state', 'error'), ('backup', self.backup.id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertEqual('One or more of the set operations failed', str(exc))

        self.volume_client.find_backup.assert_called_with(
            self.backup.id, ignore_missing=False
        )
        self.volume_client.reset_backup_status.assert_called_with(
            self.backup, status='error'
        )

    def test_backup_set_no_property(self):
        self.set_volume_api_version('3.43')

        arglist = [
            '--no-property',
            self.backup.id,
        ]
        verifylist = [
            ('no_property', True),
            ('backup', self.backup.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.volume_client.find_backup.assert_called_with(
            self.backup.id, ignore_missing=False
        )
        self.volume_client.update_backup.assert_called_once_with(
            self.backup, metadata={}
        )

    def test_backup_set_no_property_pre_v343(self):
        self.set_volume_api_version('3.42')

        arglist = [
            '--no-property',
            self.backup.id,
        ]
        verifylist = [
            ('no_property', True),
            ('backup', self.backup.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn("--os-volume-api-version 3.43 or greater", str(exc))

    def test_backup_set_property(self):
        self.set_volume_api_version('3.43')

        arglist = [
            '--property',
            'foo=bar',
            self.backup.id,
        ]
        verifylist = [
            ('properties', {'foo': 'bar'}),
            ('backup', self.backup.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.volume_client.find_backup.assert_called_with(
            self.backup.id, ignore_missing=False
        )
        self.volume_client.update_backup.assert_called_once_with(
            self.backup, metadata={'wow': 'cool', 'foo': 'bar'}
        )

    def test_backup_set_property_pre_v343(self):
        self.set_volume_api_version('3.42')

        arglist = [
            '--property',
            'foo=bar',
            self.backup.id,
        ]
        verifylist = [
            ('properties', {'foo': 'bar'}),
            ('backup', self.backup.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn("--os-volume-api-version 3.43 or greater", str(exc))


class TestBackupUnset(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.backup = sdk_fakes.generate_fake_resource(
            _backup.Backup, metadata={'foo': 'bar', 'wow': 'cool'}
        )
        self.volume_client.find_backup.return_value = self.backup
        self.volume_client.delete_backup_metadata.return_value = None

        self.cmd = volume_backup.UnsetVolumeBackup(self.app, None)

    def test_backup_unset_property(self):
        self.set_volume_api_version('3.43')

        arglist = [
            '--property',
            'foo',
            self.backup.id,
        ]
        verifylist = [
            ('properties', ['foo']),
            ('backup', self.backup.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.volume_client.find_backup.assert_called_with(
            self.backup.id, ignore_missing=False
        )
        self.volume_client.delete_backup_metadata.assert_called_once_with(
            self.backup, keys=['wow']
        )

    def test_backup_unset_property_pre_v343(self):
        self.set_volume_api_version('3.42')

        arglist = [
            '--property',
            'foo',
            self.backup.id,
        ]
        verifylist = [
            ('properties', ['foo']),
            ('backup', self.backup.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn("--os-volume-api-version 3.43 or greater", str(exc))


class TestBackupShow(volume_fakes.TestVolume):
    columns = (
        "availability_zone",
        "container",
        "created_at",
        "data_timestamp",
        "description",
        "encryption_key_id",
        "fail_reason",
        "has_dependent_backups",
        "id",
        "is_incremental",
        "metadata",
        "name",
        "object_count",
        "project_id",
        "size",
        "snapshot_id",
        "status",
        "updated_at",
        "user_id",
        "volume_id",
    )

    def setUp(self):
        super().setUp()

        self.backup = sdk_fakes.generate_fake_resource(_backup.Backup)
        self.volume_client.find_backup.return_value = self.backup

        self.data = (
            self.backup.availability_zone,
            self.backup.container,
            self.backup.created_at,
            self.backup.data_timestamp,
            self.backup.description,
            self.backup.encryption_key_id,
            self.backup.fail_reason,
            self.backup.has_dependent_backups,
            self.backup.id,
            self.backup.is_incremental,
            self.backup.metadata,
            self.backup.name,
            self.backup.object_count,
            self.backup.project_id,
            self.backup.size,
            self.backup.snapshot_id,
            self.backup.status,
            self.backup.updated_at,
            self.backup.user_id,
            self.backup.volume_id,
        )

        self.cmd = volume_backup.ShowVolumeBackup(self.app, None)

    def test_backup_show(self):
        arglist = [self.backup.id]
        verifylist = [("backup", self.backup.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_client.find_backup.assert_called_with(
            self.backup.id, ignore_missing=False
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock

from openstack.block_storage.v2 import consistency_group as _consistency_group
from openstack.block_storage.v2 import (
    consistency_group_snapshot as _consistency_group_snapshot,
)
from openstack.block_storage.v2 import type as _type
from openstack.block_storage.v2 import volume as _volume
from openstack.test import fakes as sdk_fakes
from osc_lib.cli import format_columns
from osc_lib import exceptions

from openstackclient.tests.unit.volume.v2 import fakes as volume_fakes
from openstackclient.volume.v2 import consistency_group


class TestConsistencyGroupAddVolume(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.consistency_group = sdk_fakes.generate_fake_resource(
            _consistency_group.ConsistencyGroup
        )
        self.volume_client.find_consistency_group.return_value = (
            self.consistency_group
        )
        self.cmd = consistency_group.AddVolumeToConsistencyGroup(
            self.app, None
        )

    def test_add_one_volume_to_consistency_group(self):
        volume = sdk_fakes.generate_fake_resource(_volume.Volume)
        self.volume_client.find_volume.return_value = volume
        arglist = [
            self.consistency_group.id,
            volume.id,
        ]
        verifylist = [
            ('consistency_group', self.consistency_group.id),
            ('volumes', [volume.id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.find_volume.assert_called_once_with(
            volume.id, ignore_missing=False
        )
        self.volume_client.find_consistency_group.assert_called_once_with(
            self.consistency_group.id, ignore_missing=False
        )
        self.volume_client.update_consistency_group.assert_called_once_with(
            self.consistency_group, add_volumes=volume.id
        )
        self.assertIsNone(result)

    def test_add_multiple_volumes_to_consistency_group(self):
        volumes = [
            sdk_fakes.generate_fake_resource(_volume.Volume),
            sdk_fakes.generate_fake_resource(_volume.Volume),
        ]
        self.volume_client.find_volume.side_effect = volumes
        arglist = [
            self.consistency_group.id,
            volumes[0].id,
            volumes[1].id,
        ]
        verifylist = [
            ('consistency_group', self.consistency_group.id),
            ('volumes', [volumes[0].id, volumes[1].id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.update_consistency_group.assert_called_once_with(
            self.consistency_group,
            add_volumes=volumes[0].id + ',' + volumes[1].id,
        )
        self.assertIsNone(result)

    @mock.patch.object(consistency_group.LOG, 'error')
    def test_add_multiple_volumes_to_consistency_group_with_exception(
        self, mock_error
    ):
        volume = sdk_fakes.generate_fake_resource(_volume.Volume)
        self.volume_client.find_volume.side_effect = [
            volume,
            exceptions.CommandError,
        ]
        arglist = [
            self.consistency_group.id,
            volume.id,
            'unexist_volume',
        ]
        verifylist = [
            ('consistency_group', self.consistency_group.id),
            ('volumes', [volume.id, 'unexist_volume']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        mock_error.assert_called_with(
            '%(result)s of %(total)s volumes failed to add.',
            {'result': 1, 'total': 2},
        )
        self.assertIsNone(result)
        self.volume_client.find_volume.assert_any_call(
            volume.id, ignore_missing=False
        )
        self.volume_client.find_volume.assert_any_call(
            'unexist_volume', ignore_missing=False
        )
        self.assertEqual(2, self.volume_client.find_volume.call_count)
        self.volume_client.find_consistency_group.assert_called_once_with(
            self.consistency_group.id, ignore_missing=False
        )
        self.volume_client.update_consistency_group.assert_called_once_with(
            self.consistency_group, add_volumes=volume.id
        )


class TestConsistencyGroupCreate(volume_fakes.TestVolume):
    columns = (
        'availability_zone',
        'created_at',
        'description',
        'id',
        'name',
        'status',
        'volume_types',
    )

    def setUp(self):
        super().setUp()

        self.volume_type = sdk_fakes.generate_fake_resource(_type.Type)
        self.new_consistency_group = sdk_fakes.generate_fake_resource(
            _consistency_group.ConsistencyGroup
        )
        self.consistency_group_snapshot = sdk_fakes.generate_fake_resource(
            _consistency_group_snapshot.ConsistencyGroupSnapshot
        )
        self.volume_client.create_consistency_group.return_value = (
            self.new_consistency_group
        )
        self.volume_client.create_consistency_group_from_source.return_value = self.new_consistency_group
        self.volume_client.find_type.return_value = self.volume_type
        self.volume_client.find_consistency_group_snapshot.return_value = (
            self.consistency_group_snapshot
        )
        self.volume_client.find_consistency_group.return_value = (
            self.new_consistency_group
        )

        self.data = (
            self.new_consistency_group.availability_zone,
            self.new_consistency_group.created_at,
            self.new_consistency_group.description,
            self.new_consistency_group.id,
            self.new_consistency_group.name,
            self.new_consistency_group.status,
            self.new_consistency_group.volume_types,
        )

        self.cmd = consistency_group.CreateConsistencyGroup(self.app, None)

    def test_consistency_group_create(self):
        arglist = [
            '--volume-type',
            self.volume_type.id,
            '--description',
            self.new_consistency_group.description,
            '--availability-zone',
            self.new_consistency_group.availability_zone,
            self.new_consistency_group.name,
        ]
        verifylist = [
            ('volume_type', self.volume_type.id),
            ('description', self.new_consistency_group.description),
            (
                'availability_zone',
                self.new_consistency_group.availability_zone,
            ),
            ('name', self.new_consistency_group.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.find_type.assert_called_once_with(
            self.volume_type.id, ignore_missing=False
        )
        self.volume_client.find_consistency_group.assert_not_called()
        self.volume_client.create_consistency_group.assert_called_once_with(
            volume_types=self.volume_type.id,
            name=self.new_consistency_group.name,
            description=self.new_consistency_group.description,
            availability_zone=self.new_consistency_group.availability_zone,
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_consistency_group_create_without_name(self):
        arglist = [
            '--volume-type',
            self.volume_type.id,
            '--description',
            self.new_consistency_group.description,
            '--availability-zone',
            self.new_consistency_group.availability_zone,
        ]
        verifylist = [
            ('volume_type', self.volume_type.id),
            ('description', self.new_consistency_group.description),
            (
                'availability_zone',
                self.new_consistency_group.availability_zone,
            ),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.find_type.assert_called_once_with(
            self.volume_type.id, ignore_missing=False
        )
        self.volume_client.find_consistency_group.assert_not_called()
        self.volume_client.create_consistency_group.assert_called_once_with(
            volume_types=self.volume_type.id,
            name=None,
            description=self.new_consistency_group.description,
            availability_zone=self.new_consistency_group.availability_zone,
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_consistency_group_create_from_source(self):
        arglist = [
            '--consistency-group-source',
            self.new_consistency_group.id,
            '--description',
            self.new_consistency_group.description,
            self.new_consistency_group.name,
        ]
        verifylist = [
            ('source', self.new_consistency_group.id),
            ('description', self.new_consistency_group.description),
            ('name', self.new_consistency_group.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.find_type.assert_not_called()
        self.volume_client.find_consistency_group.assert_called_once_with(
            self.new_consistency_group.id, ignore_missing=False
        )
        self.volume_client.create_consistency_group_from_source.assert_called_once_with(
            consistency_group_snapshot=None,
            consistency_group=self.new_consistency_group.id,
            name=self.new_consistency_group.name,
            description=self.new_consistency_group.description,
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_consistency_group_create_from_snapshot(self):
        arglist = [
            '--consistency-group-snapshot',
            self.consistency_group_snapshot.id,
            '--description',
            self.new_consistency_group.description,
            self.new_consistency_group.name,
        ]
        verifylist = [
            ('snapshot', self.consistency_group_snapshot.id),
            ('description', self.new_consistency_group.description),
            ('name', self.new_consistency_group.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.find_type.assert_not_called()
        self.volume_client.find_consistency_group_snapshot.assert_called_once_with(
            self.consistency_group_snapshot.id, ignore_missing=False
        )
        self.volume_client.create_consistency_group_from_source.assert_called_once_with(
            consistency_group_snapshot=self.consistency_group_snapshot.id,
            consistency_group=None,
            name=self.new_consistency_group.name,
            description=self.new_consistency_group.description,
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)


class TestConsistencyGroupDelete(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.consistency_groups = [
            sdk_fakes.generate_fake_resource(
                _consistency_group.ConsistencyGroup
            ),
            sdk_fakes.generate_fake_resource(
                _consistency_group.ConsistencyGroup
            ),
        ]
        self.volume_client.find_consistency_group.side_effect = (
            self.consistency_groups
        )
        self.volume_client.delete_consistency_group.return_value = None

        self.cmd = consistency_group.DeleteConsistencyGroup(self.app, None)

    def test_consistency_group_delete(self):
        arglist = [self.consistency_groups[0].id]
        verifylist = [("consistency_groups", [self.consistency_groups[0].id])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.find_consistency_group.assert_called_once_with(
            self.consistency_groups[0].id, ignore_missing=False
        )
        self.volume_client.delete_consistency_group.assert_called_once_with(
            self.consistency_groups[0], force=False
        )
        self.assertIsNone(result)

    def test_consistency_group_delete_with_force(self):
        arglist = [
            '--force',
            self.consistency_groups[0].id,
        ]
        verifylist = [
            ('force', True),
            ("consistency_groups", [self.consistency_groups[0].id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.delete_consistency_group.assert_called_once_with(
            self.consistency_groups[0], force=True
        )
        self.assertIsNone(result)

    def test_delete_multiple_consistency_groups(self):
        self.volume_client.find_consistency_group.side_effect = (
            self.consistency_groups
        )
        arglist = [cg.id for cg in self.consistency_groups]
        verifylist = [
            ('consistency_groups', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.volume_client.delete_consistency_group.assert_any_call(
            self.consistency_groups[0], force=False
        )
        self.volume_client.delete_consistency_group.assert_any_call(
            self.consistency_groups[1], force=False
        )
        self.assertIsNone(result)

    def test_delete_multiple_consistency_groups_with_exception(self):
        self.volume_client.find_consistency_group.side_effect = [
            self.consistency_groups[0],
            exceptions.CommandError,
        ]
        arglist = [
            self.consistency_groups[0].id,
            'unexist_consistency_group',
        ]
        verifylist = [
            ('consistency_groups', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual(
                '1 of 2 consistency groups failed to delete.', str(e)
            )

        self.volume_client.find_consistency_group.assert_any_call(
            self.consistency_groups[0].id, ignore_missing=False
        )
        self.volume_client.find_consistency_group.assert_any_call(
            'unexist_consistency_group', ignore_missing=False
        )
        self.assertEqual(
            2, self.volume_client.find_consistency_group.call_count
        )
        self.volume_client.delete_consistency_group.assert_called_once_with(
            self.consistency_groups[0], force=False
        )


class TestConsistencyGroupList(volume_fakes.TestVolume):
    column_headers = [
        'ID',
        'Status',
        'Name',
    ]
    column_headers_long = [
        'ID',
        'Status',
        'Availability Zone',
        'Name',
        'Description',
        'Volume Types',
    ]

    def setUp(self):
        super().setUp()

        self.consistency_groups = [
            sdk_fakes.generate_fake_resource(
                _consistency_group.ConsistencyGroup
            ),
            sdk_fakes.generate_fake_resource(
                _consistency_group.ConsistencyGroup
            ),
        ]
        self.volume_client.consistency_groups.return_value = (
            self.consistency_groups
        )

        self.data = []
        for c in self.consistency_groups:
            self.data.append(
                (
                    c.id,
                    c.status,
                    c.name,
                )
            )
        self.data_long = []
        for c in self.consistency_groups:
            self.data_long.append(
                (
                    c.id,
                    c.status,
                    c.availability_zone,
                    c.name,
                    c.description,
                    format_columns.ListColumn(c.volume_types),
                )
            )

        self.cmd = consistency_group.ListConsistencyGroup(self.app, None)

    def test_consistency_group_list_without_options(self):
        arglist = []
        verifylist = [
            ("all_projects", False),
            ("long", False),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.consistency_groups.assert_called_once_with(
            all_tenants=False,
        )
        self.assertEqual(self.column_headers, columns)
        self.assertCountEqual(self.data, list(data))

    def test_consistency_group_list_with_all_project(self):
        arglist = ["--all-projects"]
        verifylist = [
            ("all_projects", True),
            ("long", False),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.consistency_groups.assert_called_once_with(
            all_tenants=True,
        )
        self.assertEqual(self.column_headers, columns)
        self.assertCountEqual(self.data, list(data))

    def test_consistency_group_list_with_long(self):
        arglist = [
            "--long",
        ]
        verifylist = [
            ("all_projects", False),
            ("long", True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.consistency_groups.assert_called_once_with(
            all_tenants=False,
        )
        self.assertEqual(self.column_headers_long, columns)
        self.assertCountEqual(self.data_long, list(data))


class TestConsistencyGroupRemoveVolume(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.consistency_group = sdk_fakes.generate_fake_resource(
            _consistency_group.ConsistencyGroup
        )
        self.volume_client.find_consistency_group.return_value = (
            self.consistency_group
        )
        self.cmd = consistency_group.RemoveVolumeFromConsistencyGroup(
            self.app, None
        )

    def test_remove_one_volume_from_consistency_group(self):
        volume = sdk_fakes.generate_fake_resource(_volume.Volume)
        self.volume_client.find_volume.return_value = volume
        arglist = [
            self.consistency_group.id,
            volume.id,
        ]
        verifylist = [
            ('consistency_group', self.consistency_group.id),
            ('volumes', [volume.id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.update_consistency_group.assert_called_once_with(
            self.consistency_group, remove_volumes=volume.id
        )
        self.assertIsNone(result)

    def test_remove_multi_volumes_from_consistency_group(self):
        volumes = [
            sdk_fakes.generate_fake_resource(_volume.Volume),
            sdk_fakes.generate_fake_resource(_volume.Volume),
        ]
        self.volume_client.find_volume.side_effect = volumes
        arglist = [
            self.consistency_group.id,
            volumes[0].id,
            volumes[1].id,
        ]
        verifylist = [
            ('consistency_group', self.consistency_group.id),
            ('volumes', [volumes[0].id, volumes[1].id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.update_consistency_group.assert_called_once_with(
            self.consistency_group,
            remove_volumes=volumes[0].id + ',' + volumes[1].id,
        )
        self.assertIsNone(result)

    @mock.patch.object(consistency_group.LOG, 'error')
    def test_remove_multiple_volumes_from_consistency_group_with_exception(
        self,
        mock_error,
    ):
        volume = sdk_fakes.generate_fake_resource(_volume.Volume)
        self.volume_client.find_volume.side_effect = [
            volume,
            exceptions.CommandError,
        ]
        arglist = [
            self.consistency_group.id,
            volume.id,
            'unexist_volume',
        ]
        verifylist = [
            ('consistency_group', self.consistency_group.id),
            ('volumes', [volume.id, 'unexist_volume']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        mock_error.assert_called_with(
            '%(result)s of %(total)s volumes failed to remove.',
            {'result': 1, 'total': 2},
        )
        self.assertIsNone(result)
        self.volume_client.find_volume.assert_any_call(
            volume.id, ignore_missing=False
        )
        self.volume_client.find_volume.assert_any_call(
            'unexist_volume', ignore_missing=False
        )
        self.assertEqual(2, self.volume_client.find_volume.call_count)
        self.volume_client.find_consistency_group.assert_called_once_with(
            self.consistency_group.id, ignore_missing=False
        )
        self.volume_client.update_consistency_group.assert_called_once_with(
            self.consistency_group, remove_volumes=volume.id
        )


class TestConsistencyGroupSet(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.consistency_group = sdk_fakes.generate_fake_resource(
            _consistency_group.ConsistencyGroup
        )
        self.volume_client.find_consistency_group.return_value = (
            self.consistency_group
        )
        self.cmd = consistency_group.SetConsistencyGroup(self.app, None)

    def test_consistency_group_set_name(self):
        new_name = 'new_name'
        arglist = [
            '--name',
            new_name,
            self.consistency_group.id,
        ]
        verifylist = [
            ('name', new_name),
            ('description', None),
            ('consistency_group', self.consistency_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.find_consistency_group.assert_called_once_with(
            self.consistency_group.id, ignore_missing=False
        )
        self.volume_client.update_consistency_group.assert_called_once_with(
            self.consistency_group, name=new_name
        )
        self.assertIsNone(result)

    def test_consistency_group_set_description(self):
        new_description = 'new_description'
        arglist = [
            '--description',
            new_description,
            self.consistency_group.id,
        ]
        verifylist = [
            ('name', None),
            ('description', new_description),
            ('consistency_group', self.consistency_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.find_consistency_group.assert_called_once_with(
            self.consistency_group.id, ignore_missing=False
        )
        self.volume_client.update_consistency_group.assert_called_once_with(
            self.consistency_group, description=new_description
        )
        self.assertIsNone(result)


class TestConsistencyGroupShow(volume_fakes.TestVolume):
    columns = (
        'availability_zone',
        'created_at',
        'description',
        'id',
        'name',
        'status',
        'volume_types',
    )

    def setUp(self):
        super().setUp()

        self.consistency_group = sdk_fakes.generate_fake_resource(
            _consistency_group.ConsistencyGroup
        )
        self.data = (
            self.consistency_group.availability_zone,
            self.consistency_group.created_at,
            self.consistency_group.description,
            self.consistency_group.id,
            self.consistency_group.name,
            self.consistency_group.status,
            self.consistency_group.volume_types,
        )
        self.volume_client.find_consistency_group.return_value = (
            self.consistency_group
        )
        self.cmd = consistency_group.ShowConsistencyGroup(self.app, None)

    def test_consistency_group_show(self):
        arglist = [self.consistency_group.id]
        verifylist = [("consistency_group", self.consistency_group.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.volume_client.find_consistency_group.assert_called_once_with(
            self.consistency_group.id, ignore_missing=False
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from openstackclient.tests.functional.network.v2 import common


class NetworkRBACTests(common.NetworkTests):
    """Functional tests for network rbac"""

    OBJECT_ID: str
    ID: str
    HEADERS = ['ID']
    FIELDS = ['id']

    def setUp(self):
        super().setUp()

        self.NET_NAME = uuid.uuid4().hex
        self.PROJECT_NAME = uuid.uuid4().hex

        cmd_output = self.openstack(
            'network create ' + self.NET_NAME,
            parse_output=True,
        )
        self.addCleanup(self.openstack, 'network delete ' + cmd_output['id'])
        self.OBJECT_ID = cmd_output['id']

        cmd_output = self.openstack(
            'network rbac create '
            + self.OBJECT_ID
            + ' --action access_as_shared'
            + ' --target-project admin'
            + ' --type network',
            parse_output=True,
        )
        self.addCleanup(
            self.openstack, 'network rbac delete ' + cmd_output['id']
        )
        self.ID = cmd_output['id']
        self.assertEqual(self.OBJECT_ID, cmd_output['object_id'])

    def test_network_rbac_list(self):
        cmd_output = self.openstack('network rbac list', parse_output=True)
        self.assertIn(self.ID, [rbac['ID'] for rbac in cmd_output])

    def test_network_rbac_show(self):
        cmd_output = self.openstack(
            'network rbac show ' + self.ID,
            parse_output=True,
        )
        self.assertEqual(self.ID, cmd_output['id'])

    def test_network_rbac_set(self):
        project_id = self.openstack(
            'project create ' + self.PROJECT_NAME,
            parse_output=True,
        )['id']
        self.openstack(
            'network rbac set '
            + self.ID
            + ' --target-project '
            + self.PROJECT_NAME
        )
        cmd_output_rbac = self.openstack(
            'network rbac show ' + self.ID,
            parse_output=True,
        )
        self.assertEqual(project_id, cmd_output_rbac['target_project_id'])
        raw_output_project = self.openstack(
            'project delete ' + self.PROJECT_NAME
        )
        self.assertEqual('', raw_output_project)



# Copyright (c) 2017 Juniper networks Inc.
# All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from openstack.network.v2 import (
    bgpvpn_port_association as _bgpvpn_port_association,
)
from osc_lib.cli import format_columns
from osc_lib.cli import identity as osc_id
from osc_lib.cli import parseractions
from osc_lib import exceptions
from osc_lib import utils as osc_utils
from osc_lib.utils import columns as column_util

from openstackclient import command
from openstackclient.i18n import _

LOG = logging.getLogger(__name__)

_attr_map = (
    ('id', 'ID', column_util.LIST_BOTH),
    ('project_id', 'Project', column_util.LIST_LONG_ONLY),
    ('port_id', 'Port ID', column_util.LIST_BOTH),
    (
        'prefix_routes',
        'Prefix Routes (BGP LOCAL_PREF)',
        column_util.LIST_LONG_ONLY,
    ),
    (
        'bgpvpn_routes',
        'BGP VPN Routes (BGP LOCAL_PREF)',
        column_util.LIST_LONG_ONLY,
    ),
    (
        'advertise_fixed_ips',
        "Advertise Port's Fixed IPs",
        column_util.LIST_LONG_ONLY,
    ),
)
_formatters = {
    'prefix_routes': format_columns.ListColumn,
    'bgpvpn_routes': format_columns.ListColumn,
}


def _get_columns(
    item: _bgpvpn_port_association.BgpVpnPortAssociation,
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    column_map: dict[str, str] = {}
    hidden_columns = ['location', 'name', 'tenant_id']
    return osc_utils.get_osc_show_columns_for_sdk_resource(
        item, column_map, hidden_columns
    )


def _transform_resource(data: Any) -> Any:
    """Transforms BGP VPN port association routes property.

    Separates the two route types and formats them with ListColumn.

    {'routes':
        [
            {'type': 'prefix', 'local_pref': 100, 'prefix': '8.8.8.0/27'},
            {'type': 'bgpvpn', 'local_pref': 50,
             'bgpvpn': '157d72a9-9968-48e7-8087-6c9a9bc7a181'},
        ],
    }

    to

    {
        'prefix_routes': ['8.8.8.0/27 (100)'],
        'bgpvpn_routes': ['157d72a9-9968-48e7-8087-6c9a9bc7a181 (50)'],
    }
    """
    for route in data.get('routes', []):
        local_pref = ''
        if route.get('local_pref'):
            local_pref = ' ({local_pref})'
        if route['type'] == 'prefix':
            data.setdefault('prefix_routes', []).append(
                '{}{}'.format(route['prefix'], local_pref)
            )
        elif route['type'] == 'bgpvpn':
            data.setdefault('bgpvpn_routes', []).append(
                '{}{}'.format(route['bgpvpn_id'], local_pref)
            )
        else:
            LOG.warning("Unknown route type %s (%s).", route['type'], route)
    data.pop('routes', None)
    return data


def _get_common_parser(parser: argparse.ArgumentParser, action: str) -> None:
    """Adds to parser arguments common to create, set and unset commands.

    :params ArgumentParser parser: argparse object contains all command's
                                    arguments
    :params string action: 'create', 'set' or 'unset'
    """
    ADVERTISE_ROUTE = _(
        "Fixed IPs of the port will be advertised to the BGP VPN%s"
    ) % (_(' (default)') if action == 'create' else "")
    NOT_ADVERTISE_ROUTE = _(
        "Fixed IPs of the port will not be advertised to the BGP VPN"
    )

    LOCAL_PREF_VALUE = _(
        ". Optionally, can control the value of the BGP "
        "LOCAL_PREF of the routes that will be "
        "advertised"
    )

    ADD_PREFIX_ROUTE = (
        _("Add prefix route in CIDR notation%s") % LOCAL_PREF_VALUE
    )
    REMOVE_PREFIX_ROUTE = _("Remove prefix route in CIDR notation")
    REPEAT_PREFIX_ROUTE = _("repeat option for multiple prefix routes")

    ADD_BGVPVPN_ROUTE = (
        _("Add BGP VPN route for route leaking%s") % LOCAL_PREF_VALUE
    )
    REMOVE_BGPVPN_ROUTE = _("Remove BGP VPN route")
    REPEAT_BGPVPN_ROUTE = _("repeat option for multiple BGP VPN routes")

    group_advertise_fixed_ips = parser.add_mutually_exclusive_group()
    group_advertise_fixed_ips.add_argument(
        '--advertise-fixed-ips',
        action='store_true',
        help=NOT_ADVERTISE_ROUTE if action == 'unset' else ADVERTISE_ROUTE,
    )
    group_advertise_fixed_ips.add_argument(
        '--no-advertise-fixed-ips',
        action='store_true',
        help=ADVERTISE_ROUTE if action == 'unset' else NOT_ADVERTISE_ROUTE,
    )

    if action in ['create', 'set']:
        parser.add_argument(
            '--prefix-route',
            metavar="prefix=<cidr>[,local_pref=<integer>]",
            dest='prefix_routes',
            action=parseractions.MultiKeyValueAction,
            required_keys=['prefix'],
            optional_keys=['local_pref'],
            help=f"{ADD_PREFIX_ROUTE} ({REPEAT_PREFIX_ROUTE})",
        )
        parser.add_argument(
            '--bgpvpn-route',
            metavar="bgpvpn=<BGP VPN ID or name>[,local_pref=<integer>]",
            dest='bgpvpn_routes',
            action=parseractions.MultiKeyValueAction,
            required_keys=['bgpvpn'],
            optional_keys=['local_pref'],
            help=f"{ADD_BGVPVPN_ROUTE} ({REPEAT_BGPVPN_ROUTE})",
        )
    else:
        parser.add_argument(
            '--prefix-route',
            metavar="<cidr>",
            dest='prefix_routes',
            action='append',
            help=f"{REMOVE_PREFIX_ROUTE} ({REPEAT_PREFIX_ROUTE})",
        )
        parser.add_argument(
            '--bgpvpn-route',
            metavar="<BGP VPN ID or name>",
            dest='bgpvpn_routes',
            action='append',
            help=f"{REMOVE_BGPVPN_ROUTE} ({REPEAT_BGPVPN_ROUTE})",
        )
    if action != 'create':
        parser.add_argument(
            '--no-prefix-route' if action == 'set' else '--all-prefix-routes',
            dest='purge_prefix_route',
            action='store_true',
            help=_('Empty prefix route list'),
        )
        parser.add_argument(
            '--no-bgpvpn-route' if action == 'set' else '--all-bgpvpn-routes',
            dest='purge_bgpvpn_route',
            action='store_true',
            help=_('Empty BGP VPN route list'),
        )


def _args2body(
    client: Any, action: str, bgpvpn_id: str, args: argparse.Namespace
) -> dict[str, Any]:
    attrs: dict[str, Any] = {}

    if action != 'create':
        assoc = client.find_bgpvpn_port_association(
            args.port_association_id,
            bgpvpn_id=bgpvpn_id,
            ignore_missing=False,
        )
    else:
        assoc = {'routes': []}

    if args.advertise_fixed_ips:
        attrs['advertise_fixed_ips'] = action != 'unset'
    elif args.no_advertise_fixed_ips:
        attrs['advertise_fixed_ips'] = action == 'unset'

    prefix_routes: dict[str, Any] | None = None
    if 'purge_prefix_route' in args and args.purge_prefix_route:
        prefix_routes = {}
    else:
        prefix_routes = {
            r['prefix']: r.get('local_pref')
            for r in assoc['routes']
            if r['type'] == 'prefix'
        }
        if args.prefix_routes:
            if action in ['create', 'set']:
                prefix_routes.update(
                    {
                        r['prefix']: r.get('local_pref')
                        for r in args.prefix_routes
                    }
                )
            elif action == 'unset':
                for prefix in args.prefix_routes:
                    prefix_routes.pop(prefix, None)

    bgpvpn_routes: dict[str, Any] | None = None
    if 'purge_bgpvpn_route' in args and args.purge_bgpvpn_route:
        bgpvpn_routes = {}
    else:
        bgpvpn_routes = {
            r['bgpvpn_id']: r.get('local_pref')
            for r in assoc['routes']
            if r['type'] == 'bgpvpn'
        }
        if args.bgpvpn_routes:
            if action == 'unset':
                routes = [{'bgpvpn': bgpvpn} for bgpvpn in args.bgpvpn_routes]
            else:
                routes = args.bgpvpn_routes
            args_bgpvpn_routes = {
                client.find_bgpvpn(
                    r['bgpvpn'], ignore_missing=False
                ).id: r.get('local_pref')
                for r in routes
            }
            if action in ['create', 'set']:
                bgpvpn_routes.update(args_bgpvpn_routes)
            elif action == 'unset':
                for bgpvpn_id in args_bgpvpn_routes:
                    bgpvpn_routes.pop(bgpvpn_id, None)

    if prefix_routes is not None and not prefix_routes:
        attrs.setdefault('routes', [])
    elif prefix_routes is not None:
        for prefix, local_pref in prefix_routes.items():
            route: dict[str, Any] = {
                'type': 'prefix',
                'prefix': prefix,
            }
            if local_pref:
                route['local_pref'] = int(local_pref)
            attrs.setdefault('routes', []).append(route)
    if bgpvpn_routes is not None and not bgpvpn_routes:
        attrs.setdefault('routes', [])
    elif bgpvpn_routes is not None:
        for bgpvpn_id, local_pref in bgpvpn_routes.items():
            route = {
                'type': 'bgpvpn',
                'bgpvpn_id': bgpvpn_id,
            }
            if local_pref:
                route['local_pref'] = int(local_pref)
            attrs.setdefault('routes', []).append(route)

    return attrs


class CreateBgpvpnPortAssoc(command.ShowOne):
    _description = _("Create a BGP VPN port association")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        osc_id.add_project_owner_option_to_parser(parser)
        parser.add_argument(
            'bgpvpn',
            metavar="<bgpvpn>",
            help=_("BGP VPN to apply the port association (name or ID)"),
        )
        parser.add_argument(
            'port',
            metavar="<port>",
            help=_("Port to associate the BGP VPN (name or ID)"),
        )
        _get_common_parser(parser, 'create')
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        bgpvpn = client.find_bgpvpn(parsed_args.bgpvpn, ignore_missing=False)
        port = client.find_port(parsed_args.port, ignore_missing=False)
        body: dict[str, Any] = {'port_id': port['id']}
        if 'project' in parsed_args and parsed_args.project is not None:
            project_id = osc_id.find_project(
                self.app.client_manager.sdk_connection,
                parsed_args.project,
                parsed_args.project_domain,
            ).id
            body['project_id'] = project_id

        body.update(_args2body(client, 'create', bgpvpn['id'], parsed_args))

        obj = client.create_bgpvpn_port_association(bgpvpn['id'], **body)
        _transform_resource(obj)
        display_columns, columns = _get_columns(obj)
        data = osc_utils.get_dict_properties(
            obj, columns, formatters=_formatters
        )
        return display_columns, data


class SetBgpvpnPortAssoc(command.Command):
    _description = _("Set BGP VPN port association properties")
    _action = 'set'

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'port_association_id',
            metavar="<port association ID>",
            help=_("Port association ID to update"),
        )
        parser.add_argument(
            'bgpvpn',
            metavar="<bgpvpn>",
            help=_("BGP VPN the port association belongs to (name or ID)"),
        )
        _get_common_parser(parser, self._action)
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        bgpvpn = client.find_bgpvpn(parsed_args.bgpvpn, ignore_missing=False)
        body = _args2body(client, self._action, bgpvpn['id'], parsed_args)
        client.update_bgpvpn_port_association(
            bgpvpn['id'], parsed_args.port_association_id, **body
        )


class UnsetBgpvpnPortAssoc(SetBgpvpnPortAssoc):
    _description = _("Unset BGP VPN port association properties")
    _action = 'unset'


class DeleteBgpvpnPortAssoc(command.Command):
    _description = _(
        "Delete a BGP VPN port association(s) for a given BGP VPN"
    )

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'port_association_ids',
            metavar="<port association ID>",
            nargs="+",
            help=_("Port association ID(s) to remove"),
        )
        parser.add_argument(
            'bgpvpn',
            metavar="<bgpvpn>",
            help=_("BGP VPN the port association belongs to (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        bgpvpn = client.find_bgpvpn(parsed_args.bgpvpn, ignore_missing=False)
        fails = 0
        for id in parsed_args.port_association_ids:
            try:
                client.delete_bgpvpn_port_association(bgpvpn['id'], id)
                LOG.warning(
                    "Port association %(id)s deleted",
                    {'id': id},
                )
            except Exception as e:
                fails += 1
                LOG.error(
                    "Failed to delete port "
                    "association with ID '%(id)s': %(e)s",
                    {'id': id, 'e': e},
                )
        if fails > 0:
            msg = _(
                "Failed to delete %(fails)s of %(total)s "
                "port BGP VPN association(s)."
            ) % {
                'fails': fails,
                'total': len(parsed_args.port_association_ids),
            }
            raise exceptions.CommandError(msg)


class ListBgpvpnPortAssoc(command.Lister):
    _description = _("List BGP VPN port associations for a given BGP VPN")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'bgpvpn',
            metavar="<bgpvpn>",
            help=_("BGP VPN listed associations belong to (name or ID)"),
        )
        parser.add_argument(
            '--long',
            action='store_true',
            help=_("List additional fields in output"),
        )
        parser.add_argument(
            '--property',
            metavar="<key=value>",
            help=_(
                "Filter property to apply on returned BGP VPNs (repeat to "
                "filter on multiple properties)"
            ),
            action=parseractions.KeyValueAction,
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        client = self.app.client_manager.network
        bgpvpn = client.find_bgpvpn(parsed_args.bgpvpn, ignore_missing=False)
        params = {}
        if parsed_args.property:
            params.update(parsed_args.property)
        objs = client.bgpvpn_port_associations(
            bgpvpn['id'], retrieve_all=True, **params
        )
        transformed_objs = [_transform_resource(obj) for obj in objs]
        headers, columns = column_util.get_column_definitions(
            list(_attr_map), long_listing=parsed_args.long
        )
        return (
            headers,
            (
                osc_utils.get_dict_properties(
                    s, columns, formatters=_formatters
                )
                for s in transformed_objs
            ),
        )


class ShowBgpvpnPortAssoc(command.ShowOne):
    _description = _("Show information of a given BGP VPN port association")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'port_association_id',
            metavar="<port association ID>",
            help=_("Port association ID to look up"),
        )
        parser.add_argument(
            'bgpvpn',
            metavar="<bgpvpn>",
            help=_("BGP VPN the association belongs to (name or ID)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        bgpvpn = client.find_bgpvpn(parsed_args.bgpvpn, ignore_missing=False)
        obj = client.get_bgpvpn_port_association(
            bgpvpn['id'], parsed_args.port_association_id
        )
        _transform_resource(obj)
        display_columns, columns = _get_columns(obj)
        data = osc_utils.get_dict_properties(
            obj, columns, formatters=_formatters
        )
        return display_columns, data



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Network agent action implementations"""

import argparse
import logging
from collections.abc import Iterable, Sequence
from typing import Any

from cliff import columns as cliff_columns
from openstack.network.v2 import agent as _agent
from osc_lib.cli import format_columns
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.common import pagination
from openstackclient.i18n import _

LOG = logging.getLogger(__name__)


class AliveColumn(cliff_columns.FormattableColumn[bool]):
    def human_readable(self) -> str:
        return ":-)" if self._value else "XXX"


class AdminStateColumn(cliff_columns.FormattableColumn[bool]):
    def human_readable(self) -> str:
        return 'UP' if self._value else 'DOWN'


_formatters = {
    'is_alive': AliveColumn,
    'alive': AliveColumn,
    'admin_state_up': AdminStateColumn,
    'is_admin_state_up': AdminStateColumn,
    'configurations': format_columns.DictColumn,
}


def _get_network_columns(
    item: _agent.Agent,
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    column_data_mapping = {
        'admin_state_up': 'is_admin_state_up',
        'agent_type': 'agent_type',
        'alive': 'is_alive',
        'availability_zone': 'availability_zone',
        'binary': 'binary',
        'configuration': 'configuration',
        'created_at': 'created_at',
        'description': 'description',
        'ha_chassis_priority': 'ha_chassis_priority',
        'ha_state': 'ha_state',
        'host': 'host',
        'id': 'id',
        'last_heartbeat_at': 'last_heartbeat_at',
        'resources_synced': 'resources_synced',
        'started_at': 'started_at',
        'topic': 'topic',
    }
    return (
        tuple(column_data_mapping.keys()),
        tuple(column_data_mapping.values()),
    )


class AddNetworkToAgent(command.Command):
    _description = _("Add network to an agent")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--dhcp',
            action='store_true',
            help=_('Add network to a DHCP agent'),
        )
        parser.add_argument(
            'agent_id',
            metavar='<agent-id>',
            help=_('Agent to which a network is added (ID only)'),
        )
        parser.add_argument(
            'network',
            metavar='<network>',
            help=_('Network to be added to an agent (name or ID)'),
        )

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        agent = client.get_agent(parsed_args.agent_id)
        network = client.find_network(
            parsed_args.network, ignore_missing=False
        )
        if parsed_args.dhcp:
            try:
                client.add_dhcp_agent_to_network(agent, network)
            except Exception:
                msg = f'Failed to add {network.name} to {agent.agent_type}'
                exceptions.CommandError(msg)


class AddRouterToAgent(command.Command):
    _description = _("Add router to an agent")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--l3', action='store_true', help=_('Add router to an L3 agent')
        )
        parser.add_argument(
            '--ha-chassis-priority',
            metavar='<ha-chassis-priority>',
            type=int,
            help=_(
                "HA Chassis priority, ranging from [0, 32767]. "
                "Only used with --l3 and for ML2/OVN L3 agents"
            ),
        )
        parser.add_argument(
            'agent_id',
            metavar='<agent-id>',
            help=_("Agent to which a router is added (ID only)"),
        )
        parser.add_argument(
            'router',
            metavar='<router>',
            help=_("Router to be added to an agent (name or ID)"),
        )

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        agent = client.get_agent(parsed_args.agent_id)
        router = client.find_router(parsed_args.router, ignore_missing=False)
        if parsed_args.l3:
            client.add_router_to_agent(
                agent,
                router,
                ha_chassis_priority=parsed_args.ha_chassis_priority,
            )


class DeleteNetworkAgent(command.Command):
    _description = _("Delete network agent(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'network_agent',
            metavar="<network-agent>",
            nargs='+',
            help=(_("Network agent(s) to delete (ID only)")),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        result = 0

        for agent in parsed_args.network_agent:
            try:
                client.delete_agent(agent, ignore_missing=False)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete network agent with "
                        "ID '%(agent)s': %(e)s"
                    ),
                    {'agent': agent, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.network_agent)
            msg = _(
                "%(result)s of %(total)s network agents failed to delete."
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class ListNetworkAgent(command.Lister):
    _description = _("List network agents")
    _supported_agents = {
        'bgp': 'BGP dynamic routing agent',
        'dhcp': 'DHCP agent',
        'open-vswitch': 'Open vSwitch agent',
        'linux-bridge': 'Linux bridge agent',
        'ofa': 'OFA driver agent',
        'l3': 'L3 agent',
        'loadbalancer': 'Loadbalancer agent',
        'metering': 'Metering agent',
        'metadata': 'Metadata agent',
        'macvtap': 'Macvtap agent',
        'nic': 'NIC Switch agent',
        'baremetal': 'Baremetal Node',
        'ovn-controller': 'OVN Controller agent',
        'ovn-controller-gateway': 'OVN Controller Gateway agent',
        'ovn-metadata': 'OVN Metadata agent',
        'ovn-agent': 'OVN Neutron agent',
    }

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        supported_agents = ','.join(self._supported_agents.keys())
        parser.add_argument(
            '--agent-type',
            metavar='<agent-type>',
            choices=list(self._supported_agents.keys()),
            help=_(
                "List only agents with the specified agent type. "
                "The supported agent types are: %(supported_agents)s."
            )
            % {'supported_agents': supported_agents},
        )
        parser.add_argument(
            '--host',
            metavar='<host>',
            help=_("List only agents running on the specified host"),
        )
        agent_type_group = parser.add_mutually_exclusive_group()
        agent_type_group.add_argument(
            '--network',
            metavar='<network>',
            help=_('List agents hosting the specified network (name or ID)'),
        )
        agent_type_group.add_argument(
            '--router',
            metavar='<router>',
            help=_('List agents hosting the specified router (name or ID)'),
        )
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_("List additional fields in output"),
        )
        pagination.add_marker_pagination_option_to_parser(parser)

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        client = self.app.client_manager.network
        columns: tuple[str, ...] = (
            'id',
            'agent_type',
            'host',
            'availability_zone',
            'is_alive',
            'is_admin_state_up',
            'binary',
        )
        column_headers: tuple[str, ...] = (
            'ID',
            'Agent Type',
            'Host',
            'Availability Zone',
            'Alive',
            'State',
            'Binary',
        )

        filters = {}
        if parsed_args.marker is not None:
            filters['marker'] = parsed_args.marker
        if parsed_args.limit is not None:
            filters['limit'] = parsed_args.limit
        if parsed_args.max_items is not None:
            filters['max_items'] = parsed_args.max_items

        data: list[_agent.Agent]
        if parsed_args.network is not None:
            network = client.find_network(
                parsed_args.network, ignore_missing=False
            )
            data = list(client.network_hosting_dhcp_agents(network))
        elif parsed_args.router is not None:
            if parsed_args.long:
                columns += ('ha_state',)
                column_headers += ('HA State',)
            router = client.find_router(
                parsed_args.router, ignore_missing=False
            )
            data = list(client.routers_hosting_l3_agents(router))
        else:
            if parsed_args.agent_type is not None:
                filters['agent_type'] = self._supported_agents[
                    parsed_args.agent_type
                ]
            if parsed_args.host is not None:
                filters['host'] = parsed_args.host

            data = list(client.agents(**filters))

        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters=_formatters,
                )
                for s in data
            ),
        )


class RemoveNetworkFromAgent(command.Command):
    _description = _("Remove network from an agent.")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--dhcp',
            action='store_true',
            help=_('Remove network from DHCP agent'),
        )
        parser.add_argument(
            'agent_id',
            metavar='<agent-id>',
            help=_('Agent to which a network is removed (ID only)'),
        )
        parser.add_argument(
            'network',
            metavar='<network>',
            help=_('Network to be removed from an agent (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        agent = client.get_agent(parsed_args.agent_id)
        network = client.find_network(
            parsed_args.network, ignore_missing=False
        )
        if parsed_args.dhcp:
            try:
                client.remove_dhcp_agent_from_network(agent, network)
            except Exception:
                msg = f'Failed to remove {network.name} to {agent.agent_type}'
                exceptions.CommandError(msg)


class RemoveRouterFromAgent(command.Command):
    _description = _("Remove router from an agent")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--l3',
            action='store_true',
            help=_('Remove router from an L3 agent'),
        )
        parser.add_argument(
            'agent_id',
            metavar='<agent-id>',
            help=_("Agent from which router will be removed (ID only)"),
        )
        parser.add_argument(
            'router',
            metavar='<router>',
            help=_("Router to be removed from an agent (name or ID)"),
        )

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        agent = client.get_agent(parsed_args.agent_id)
        router = client.find_router(parsed_args.router, ignore_missing=False)
        if parsed_args.l3:
            client.remove_router_from_agent(agent, router)


class SetNetworkAgentRouter(command.Command):
    _description = _("Set properties of a router associated to an agent")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--ha-chassis-priority',
            metavar='<ha-chassis-priority>',
            type=int,
            required=True,
            help=_(
                "HA Chassis priority, ranging from [0, 32767]. "
                "Only used with --l3 and for ML2/OVN L3 agents"
            ),
        )
        parser.add_argument(
            'agent_id',
            metavar='<agent-id>',
            help=_("L3 agent to modify (ID only)"),
        )
        parser.add_argument(
            'router',
            metavar='<router>',
            help=_("Router to update (name or ID)"),
        )

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        agent = client.get_agent(parsed_args.agent_id)
        router = client.find_router(parsed_args.router, ignore_missing=False)
        client.update_router_in_agent(
            agent,
            router,
            ha_chassis_priority=parsed_args.ha_chassis_priority,
        )


class SetNetworkAgent(command.Command):
    _description = _("Set network agent properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'network_agent',
            metavar="<network-agent>",
            help=(_("Network agent to modify (ID only)")),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_("Set network agent description"),
        )
        admin_group = parser.add_mutually_exclusive_group()
        admin_group.add_argument(
            '--enable', action='store_true', help=_("Enable network agent")
        )
        admin_group.add_argument(
            '--disable', action='store_true', help=_("Disable network agent")
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        obj = client.get_agent(parsed_args.network_agent)
        attrs = {}
        if parsed_args.description is not None:
            attrs['description'] = parsed_args.description
        if parsed_args.enable:
            attrs['is_admin_state_up'] = True
            attrs['admin_state_up'] = True
        if parsed_args.disable:
            attrs['is_admin_state_up'] = False
            attrs['admin_state_up'] = False
        client.update_agent(obj, **attrs)


class ShowNetworkAgent(command.ShowOne):
    _description = _("Display network agent details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'network_agent',
            metavar="<network-agent>",
            help=(_("Network agent to display (ID only)")),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        obj = client.get_agent(parsed_args.network_agent)
        display_columns, columns = _get_network_columns(obj)
        data = utils.get_item_properties(
            obj,
            columns,
            formatters=_formatters,
        )
        return display_columns, data
